from gwproactor.actors.actor import Actor
from gwproactor.actors.actor import SyncThreadActor
from gwproactor.actors.actor import SyncThreadT


__all__ = [
    "Actor",
    "SyncThreadT",
    "SyncThreadActor",
]



from gwproactor_test.dummies.child.config import DummyChildSettings
from gwproactor_test.dummies.child.dummy import DummyChild
from gwproactor_test.dummies.names import DUMMY_CHILD_ENV_PREFIX
from gwproactor_test.dummies.names import DUMMY_CHILD_NAME
from gwproactor_test.dummies.names import DUMMY_ENV_PREFIX
from gwproactor_test.dummies.names import DUMMY_PARENT_ENV_PREFIX
from gwproactor_test.dummies.names import DUMMY_PARENT_NAME
from gwproactor_test.dummies.parent.config import DummyParentSettings
from gwproactor_test.dummies.parent.dummy import DummyParent


__all__ = [
    "DUMMY_CHILD_NAME",
    "DUMMY_PARENT_NAME",
    "DUMMY_ENV_PREFIX",
    "DUMMY_CHILD_ENV_PREFIX",
    "DUMMY_PARENT_ENV_PREFIX",
    "DummyParent",
    "DummyParentSettings",
    "DummyChild",
    "DummyChildSettings",
]



import dataclasses
from abc import abstractmethod
from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import Callable
from typing import Optional
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import cast

from gwproto import Message
from gwproto.messages import CommEvent
from gwproto.messages import EventT
from gwproto.messages import PingMessage
from paho.mqtt.client import MQTT_ERR_SUCCESS
from paho.mqtt.client import MQTTMessageInfo

from gwproactor import Proactor
from gwproactor import ProactorSettings
from gwproactor import Runnable
from gwproactor import ServicesInterface
from gwproactor.config import LoggerLevels
from gwproactor.links import LinkManager
from gwproactor.links import MQTTClients
from gwproactor.links import MQTTClientWrapper
from gwproactor.message import DBGCommands
from gwproactor.message import DBGPayload
from gwproactor.message import MQTTReceiptPayload
from gwproactor.message import MQTTSubackPayload
from gwproactor.stats import LinkStats
from gwproactor.stats import ProactorStats


def split_subscriptions(client_wrapper: MQTTClientWrapper) -> Tuple[int, Optional[int]]:
    for i, (topic, qos) in enumerate(client_wrapper.subscription_items()):
        MQTTClientWrapper.subscribe(client_wrapper, topic, qos)
    return MQTT_ERR_SUCCESS, None


@dataclass
class RecorderLinkStats(LinkStats):
    comm_events: list[CommEvent] = field(default_factory=list)

    def __str__(self) -> str:
        s = super().__str__()
        if self.comm_events:
            s += "\n  Comm events:"
            for comm_event in self.comm_events:
                s += f"\n    {comm_event}"
        return s


class RecorderStats(ProactorStats):
    @classmethod
    def make_link(cls, link_name: str) -> RecorderLinkStats:
        return RecorderLinkStats(link_name)


ProactorT = TypeVar("ProactorT", bound=Proactor)


class RecorderInterface(ServicesInterface, Runnable):
    @classmethod
    @abstractmethod
    def make_stats(cls) -> RecorderStats: ...

    @abstractmethod
    def split_client_subacks(self, client_name: str): ...

    @abstractmethod
    def restore_client_subacks(self, client_name: str): ...

    @abstractmethod
    def pause_subacks(self): ...

    @abstractmethod
    def release_subacks(self, num_released: int = -1): ...

    @abstractmethod
    def ping_peer(self): ...

    @abstractmethod
    def summary_str(self): ...

    @abstractmethod
    def summarize(self): ...

    @property
    @abstractmethod
    def mqtt_clients(self) -> MQTTClients: ...

    @abstractmethod
    def mqtt_client_wrapper(self, client_name: str) -> MQTTClientWrapper: ...

    @abstractmethod
    def mqtt_subscriptions(self, client_name: str) -> list[str]: ...

    @abstractmethod
    def disable_derived_events(self) -> None: ...

    @abstractmethod
    def enable_derived_events(self) -> None: ...

    @abstractmethod
    def mqtt_quiescent(self) -> bool: ...


@dataclass
class _PausedAck:
    client: str
    message: Message
    qos: int
    context: Optional[Any]


class RecorderLinks(LinkManager):
    acks_paused: bool
    needs_ack: list[_PausedAck]

    # noinspection PyMissingConstructor
    def __init__(self, other: LinkManager):
        self.__dict__ = other.__dict__
        self.acks_paused = False
        self.needs_ack = []

    def publish_message(
        self, client, message: Message, qos: int = 0, context: Any = None
    ) -> MQTTMessageInfo:
        if self.acks_paused:
            self.needs_ack.append(_PausedAck(client, message, qos, context))
            return MQTTMessageInfo(-1)
        else:
            # noinspection PyProtectedMember
            return super().publish_message(client, message, qos=qos, context=context)

    def release_acks(self, clear: bool = False, num_to_release: int = -1) -> int:
        # self._logger.info(
        #     f"++release_acks: clear:{clear}  num_to_release:{num_to_release}"
        # )
        # path_dbg = 0
        if clear or num_to_release < 1:
            # path_dbg |= 0x00000001
            self.acks_paused = False
            needs_ack = self.needs_ack
            self.needs_ack = []
        else:
            # path_dbg |= 0x00000002
            num_to_release = min(num_to_release, len(self.needs_ack))
            needs_ack = self.needs_ack[:num_to_release]
            self.needs_ack = self.needs_ack[num_to_release:]
            # self._logger.info(f"needs_ack: {needs_ack}")
            # self._logger.info(f"self.needs_ack: {self.needs_ack}")
        if not clear:
            # path_dbg |= 0x00000004
            for paused_ack in needs_ack:
                # path_dbg |= 0x00000008
                super().publish_message(**dataclasses.asdict(paused_ack))
        # self._logger.info(
        #     f"--release_acks: clear:{clear}  num_to_release:{num_to_release}  path:0x{path_dbg:08X}"
        # )
        return len(needs_ack)

    def generate_event(self, event: EventT) -> None:
        if isinstance(event, CommEvent):
            cast(
                RecorderLinkStats, self._stats.link(event.PeerName)
            ).comm_events.append(event)
        super().generate_event(event)


def make_recorder_class(
    proactor_type: Type[ProactorT],
) -> Callable[..., RecorderInterface]:
    class Recorder(proactor_type):
        subacks_paused: bool
        pending_subacks: list[Message]
        mqtt_messages_dropped: bool

        def __init__(self, name: str, settings: ProactorSettings, **kwargs_):
            super().__init__(name=name, settings=settings, **kwargs_)
            self.subacks_paused = False
            self.pending_subacks = []
            self.mqtt_messages_dropped = False
            self._links = RecorderLinks(self._links)

        @classmethod
        def make_stats(cls) -> RecorderStats:
            return RecorderStats()

        @property
        def needs_ack(self) -> list[_PausedAck]:
            return self._links.needs_ack

        def split_client_subacks(self: ProactorT, client_name: str):
            client_wrapper = self.mqtt_client_wrapper(client_name)

            def member_split_subscriptions():
                return split_subscriptions(client_wrapper)

            client_wrapper.subscribe_all = member_split_subscriptions

        def restore_client_subacks(self: ProactorT, client_name: str):
            client_wrapper = self.mqtt_client_wrapper(client_name)
            client_wrapper.subscribe_all = MQTTClientWrapper.subscribe_all

        def pause_subacks(self):
            self.subacks_paused = True

        def release_subacks(self: ProactorT, num_released: int = -1):
            self.subacks_paused = False
            if num_released < 0:
                num_released = len(self.pending_subacks)
            release = self.pending_subacks[:num_released]
            self.pending_subacks = self.pending_subacks[num_released:]
            for message in release:
                self._receive_queue.put_nowait(message)

        async def process_message(self, message: Message):
            if self.subacks_paused and isinstance(message.Payload, MQTTSubackPayload):
                self.pending_subacks.append(message)
            else:
                await super().process_message(message)

        def pause_acks(self):
            self._links.acks_paused = True

        def release_acks(self, clear: bool = False, num_to_release: int = -1) -> int:
            return self._links.release_acks(clear, num_to_release=num_to_release)

        def set_ack_timeout_seconds(self, delay: float) -> None:
            self._links._acks._default_delay_seconds = delay  # noqa

        def drop_mqtt(self, drop: bool):
            self.mqtt_messages_dropped = drop

        def _process_mqtt_message(self, message: Message[MQTTReceiptPayload]):
            if not self.mqtt_messages_dropped:
                # noinspection PyProtectedMember
                super()._process_mqtt_message(message)

        def summary_str(self: ProactorT) -> str:
            s = str(self.stats)
            s += "\nLink states:\n"
            for link_name in self.stats.links:
                s += f"  {link_name:10s}  {self._links.link_state(link_name).value}\n"
            s += "Pending acks:\n"
            for link_name in self.stats.links:
                s += f"  {link_name:10s}  {self._links.num_acks(link_name):3d}\n"
            s += self._links.get_reuploads_str() + "\n"
            s += f"subacks_paused: {self.subacks_paused}  pending_subacks: {len(self.pending_subacks)}\n"
            return s

        def summarize(self: ProactorT) -> None:
            self._logger.info(self.summary_str())

        def ping_peer(self) -> None:
            self._links.publish_message(
                self.primary_peer_client, PingMessage(Src=self.publication_name)
            )

        @property
        def mqtt_clients(self) -> MQTTClients:
            return self._links.mqtt_clients()

        def mqtt_client_wrapper(self, client_name: str) -> MQTTClientWrapper:
            return self._links.mqtt_client_wrapper(client_name)

        def mqtt_subscriptions(self, client_name: str) -> list[str]:
            return [
                item[0]
                for item in self.mqtt_client_wrapper(client_name).subscription_items()
            ]

        def send_dbg_to_peer(
            self,
            message_summary: int = -1,
            lifecycle: int = -1,
            comm_event: int = -1,
            command: Optional[DBGCommands | str] = None,
        ):
            if isinstance(command, str):
                command = DBGCommands(command)
            self.send_threadsafe(
                Message(
                    Src=self.name,
                    Dst=self.name,
                    Payload=DBGPayload(
                        Levels=LoggerLevels(
                            message_summary=message_summary,
                            lifecycle=lifecycle,
                            comm_event=comm_event,
                        ),
                        Command=command,
                    ),
                )
            )

        def _derived_process_message(self, message: Message):
            match message.Payload:
                case DBGPayload():
                    message.Header.Src = self.publication_name
                    message.Header.Dst = self.primary_peer_client
                    self._links.publish_message(self.primary_peer_client, message)
                case _:
                    # noinspection PyProtectedMember
                    super()._derived_process_message(message)

        def mqtt_quiescent(self) -> bool:
            if hasattr(super(), "mqtt_quiescent"):
                return getattr(super(), "mqtt_quiescent")()
            return self._links.link(self.upstream_client).active_for_send()

        def _call_super_if_present(self, function_name: str) -> None:
            if hasattr(super(), function_name):
                getattr(super(), function_name)()

        def disable_derived_events(self) -> None:
            self._call_super_if_present("disable_dervived_events")

        def enable_derived_events(self) -> None:
            self._call_super_if_present("enable_dervived_events")

    return Recorder



import logging
from typing import Any
from typing import Optional

import pendulum


class MessageSummary:
    """Helper class for formating message summaries message receipt/publication single line summaries."""

    DEFAULT_FORMAT = (
        "  {direction:15s}  {actor_alias:40s}  {broker_flag}  {arrow:2s}  {topic:90s}"
        "  {payload_type:40s}  {message_id}"
    )

    @classmethod
    def format(
        cls,
        direction: str,
        actor_alias: str,
        topic: str,
        payload_object: Any = None,
        broker_flag: str = " ",
        timestamp: Optional[pendulum.datetime] = None,
        include_timestamp: bool = False,
        message_id: str = "",
    ) -> str:
        """
        Formats a single line summary of message receipt/publication.

        Args:
            direction: "IN" or "OUT"
            actor_alias: The node alias of the sending or receiving actor.
            topic: The destination or source topic.
            payload_object: The payload of the message.
            broker_flag: "*" for the "gw" broker.
            timestamp: pendulum.now("UTC") by default.
            include_timestamp: whether timestamp is prepended to output.
            message_id: The message id. Ignored if empty.

        Returns:
            Formatted string.
        """
        try:
            if timestamp is None:
                timestamp = pendulum.now("UTC")
            if include_timestamp:
                format_ = "{timestamp}  " + cls.DEFAULT_FORMAT
            else:
                format_ = cls.DEFAULT_FORMAT
            direction = direction.strip()
            if direction.startswith("OUT") or direction.startswith("SND"):
                arrow = "->"
            elif direction.startswith("IN") or direction.startswith("RCV"):
                arrow = "<-"
            else:
                arrow = "? "
            if hasattr(payload_object, "payload"):
                payload_object = payload_object.payload
            if hasattr(payload_object, "__class__"):
                payload_str = payload_object.__class__.__name__
            else:
                payload_str = type(payload_object)
            if message_id:
                if len(message_id) > 11:
                    message_id = f"{message_id[:8]}..."
            return format_.format(
                timestamp=timestamp.isoformat(),
                direction=direction,
                actor_alias=actor_alias,
                broker_flag=broker_flag,
                arrow=arrow,
                topic=f"[{topic}]",
                payload_type=payload_str,
                message_id=message_id,
            )
        except Exception as e:
            print(f"ouch got {e}")
            return ""


class ProactorLogger(logging.LoggerAdapter):
    MESSAGE_DELIMITER_WIDTH = 88
    MESSAGE_ENTRY_DELIMITER = "+" * MESSAGE_DELIMITER_WIDTH
    MESSAGE_EXIT_DELIMITER = "-" * MESSAGE_DELIMITER_WIDTH

    message_summary_logger: logging.Logger
    lifecycle_logger: logging.Logger
    comm_event_logger: logging.Logger

    def __init__(
        self,
        base: str,
        message_summary: str,
        lifecycle: str,
        comm_event: str,
        extra: Optional[dict] = None,
    ):
        super().__init__(logging.getLogger(base), extra=extra)
        self.message_summary_logger = logging.getLogger(message_summary)
        self.lifecycle_logger = logging.getLogger(lifecycle)
        self.comm_event_logger = logging.getLogger(comm_event)

    @property
    def general_enabled(self) -> bool:
        return self.logger.isEnabledFor(logging.INFO)

    @property
    def message_summary_enabled(self) -> bool:
        return self.message_summary_logger.isEnabledFor(logging.INFO)

    @property
    def path_enabled(self) -> bool:
        return self.message_summary_logger.isEnabledFor(logging.DEBUG)

    @property
    def lifecycle_enabled(self) -> bool:
        return self.lifecycle_logger.isEnabledFor(logging.INFO)

    @property
    def comm_event_enabled(self) -> bool:
        return self.comm_event_logger.isEnabledFor(logging.INFO)

    def message_summary(
        self,
        direction: str,
        actor_alias: str,
        topic: str,
        payload_object: Any = None,
        broker_flag: str = " ",
        timestamp: Optional[pendulum.datetime] = None,
        message_id: str = "",
    ) -> None:
        if self.message_summary_logger.isEnabledFor(logging.INFO):
            self.message_summary_logger.info(
                MessageSummary.format(
                    direction=direction,
                    actor_alias=actor_alias,
                    topic=topic,
                    payload_object=payload_object,
                    broker_flag=broker_flag,
                    timestamp=timestamp,
                    message_id=message_id,
                )
            )

    def path(self, msg: str, *args, **kwargs) -> None:
        self.message_summary_logger.debug(msg, *args, **kwargs)

    def lifecycle(self, msg: str, *args, **kwargs) -> None:
        self.lifecycle_logger.info(msg, *args, **kwargs)

    def comm_event(self, msg: str, *args, **kwargs) -> None:
        self.comm_event_logger.info(msg, *args, **kwargs)

    def message_enter(self, msg: str, *args, **kwargs) -> None:
        if self.path_enabled:
            self.path("")
            self.path(self.MESSAGE_ENTRY_DELIMITER)
            self.path(msg, *args, **kwargs)

    def message_exit(self, msg: str, *args, **kwargs) -> None:
        if self.path_enabled:
            self.path(msg, *args, **kwargs)
            self.path(self.MESSAGE_EXIT_DELIMITER)

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} "
            f"{self.logger.name}, "
            f"{self.message_summary_logger.name}, "
            f"{self.lifecycle_logger.name}, "
            f"{self.comm_event_logger.name}>"
        )



from pydantic import BaseModel
from pydantic import BaseSettings
from pydantic import root_validator
from pydantic import validator

from gwproactor.config.logging import LoggingSettings
from gwproactor.config.mqtt import MQTTClient
from gwproactor.config.paths import Paths


MQTT_LINK_POLL_SECONDS = 60.0
ACK_TIMEOUT_SECONDS = 5.0
NUM_INITIAL_EVENT_REUPLOADS: int = 5


class ProactorSettings(BaseSettings):
    paths: Paths = None
    logging: LoggingSettings = LoggingSettings()
    mqtt_link_poll_seconds: float = MQTT_LINK_POLL_SECONDS
    ack_timeout_seconds: float = ACK_TIMEOUT_SECONDS
    num_initial_event_reuploads: int = NUM_INITIAL_EVENT_REUPLOADS

    class Config:
        env_prefix = "PROACTOR_"
        env_nested_delimiter = "__"

    @validator("paths", always=True)
    def get_paths(cls, v: Paths) -> Paths:
        if v is None:
            v = Paths()
        return v

    @classmethod
    def update_paths_name(cls, values: dict, name: str) -> dict:
        """Update paths member with a new 'name' attribute, e.g., a name known by a derived class.

        This is meant to be called in a 'pre=True' root validator of a derived class.
        """
        if "paths" not in values:
            values["paths"] = Paths(name=name)
        else:
            if isinstance(values["paths"], BaseModel):
                if "name" not in values["paths"].__fields_set__:
                    values["paths"] = values["paths"].copy(name=name, deep=True)
            else:
                if "name" not in values["paths"]:
                    values["paths"]["name"] = name
        return values

    @root_validator(skip_on_failure=True)
    def post_root_validator(cls, values: dict) -> dict:
        """Update unset paths of any member MQTTClient's TLS paths based on ProactorSettings 'paths' member."""
        if not isinstance(values["paths"], Paths):
            raise ValueError(
                f"ERROR. 'paths' member must be instance of Paths. Got: {type(values['paths'])}"
            )
        for k, v in values.items():
            if isinstance(v, MQTTClient):
                v.update_tls_paths(values["paths"].certs_dir, k)
        return values



"""Classes providing interaction between synchronous and asynchronous code"""

import asyncio
import queue
import threading
import time
import traceback
from abc import ABC
from typing import Any
from typing import Optional

from gwproactor.message import InternalShutdownMessage
from gwproactor.message import PatInternalWatchdogMessage


DEFAULT_STEP_DURATION = 0.1

JOIN_NEVER_TIMEOUT = 1000000000.0  # > 30 years
JOIN_CHECK_THREAD_SECONDS = 1.0


async def async_polling_thread_join(
    t: Optional[threading.Thread],
    timeout_seconds: float = JOIN_NEVER_TIMEOUT,
    check_thread_seconds: float = JOIN_CHECK_THREAD_SECONDS,
    raise_errors: bool = False,
) -> Optional[BaseException]:
    if t is None:
        return
    if timeout_seconds is None:
        timeout_seconds = JOIN_NEVER_TIMEOUT
    end = time.time() + timeout_seconds
    remaining = timeout_seconds
    returned_e: Optional[BaseException] = None
    try:
        while t.is_alive() and remaining > 0:
            await asyncio.sleep(min(check_thread_seconds, remaining))
            remaining = end - time.time()
    except BaseException as e:
        returned_e = e
        if raise_errors:
            raise e
    return returned_e


def responsive_sleep(
    obj,
    seconds: float,
    step_duration: float = DEFAULT_STEP_DURATION,
    running_field_name: str = "_main_loop_running",
    running_field: bool = True,
) -> bool:
    """Sleep in way that is more responsive to thread termination: sleep in step_duration increments up to
    specificed seconds, at after each step checking obj._main_loop_running. If the designated running_field_name actually
    indicates that a stop has been requested (e.g. what you would expect from a field named '_stop_requested'),
    set running_field parameter to False."""
    end_time = time.time() + seconds
    while (
        getattr(obj, running_field_name) == running_field
        and (now := time.time()) < end_time
    ):
        time.sleep(min(end_time - now, step_duration))
    return getattr(obj, running_field_name) == running_field


class AsyncQueueWriter:
    """Allow synchronous code to write to an asyncio Queue.

    It is assumed the asynchronous reader has access to the asyncio Queue "await get()" from directly from it.
    """

    _loop: Optional[asyncio.AbstractEventLoop] = None
    _async_queue: Optional[asyncio.Queue] = None

    def set_async_loop(
        self, loop: asyncio.AbstractEventLoop, async_queue: asyncio.Queue
    ) -> None:
        self._loop = loop
        self._async_queue = async_queue

    def put(self, item: Any) -> None:
        """Write to asyncio queue in a threadsafe way."""
        if self._loop is None or self._async_queue is None:
            raise ValueError(
                "ERROR. start(loop, async_queue) must be called prior to put(item)"
            )
        self._loop.call_soon_threadsafe(self._async_queue.put_nowait, item)


class SyncAsyncQueueWriter:
    """Provide a full duplex communication "channel" between synchronous and asynchronous code.

    It is assumed the asynchronous reader has access to the asyncio Queue "await get()" from directly from it.
    """

    _loop: Optional[asyncio.AbstractEventLoop] = None
    _async_queue: Optional[asyncio.Queue] = None
    sync_queue: Optional[queue.Queue]

    def __init__(self, sync_queue: Optional[queue.Queue] = None):
        self.sync_queue = sync_queue

    def set_async_loop(
        self, loop: asyncio.AbstractEventLoop, async_queue: asyncio.Queue
    ) -> None:
        self._loop = loop
        self._async_queue = async_queue

    def put_to_sync_queue(
        self, item: Any, block: bool = True, timeout: Optional[float] = None
    ):
        """Write to synchronous queue in a threadsafe way."""
        self.sync_queue.put(item, block=block, timeout=timeout)

    def put_to_async_queue(self, item: Any):
        """Write to asynchronous queue in a threadsafe way."""
        if self._loop is None or self._async_queue is None:
            raise ValueError(
                "ERROR. start(loop, async_queue) must be called prior to put(item)"
            )
        self._loop.call_soon_threadsafe(self._async_queue.put_nowait, item)

    def get_from_sync_queue(
        self, block: bool = True, timeout: Optional[float] = None
    ) -> Any:
        """Read from synchronous queue in a threadsafe way."""
        return self.sync_queue.get(block=block, timeout=timeout)


class SyncAsyncInteractionThread(threading.Thread, ABC):
    """A thread wrapper providing an async-sync communication channel and simple "iterate, sleep, read message"
    semantics.
    """

    SLEEP_STEP_SECONDS = 0.1
    PAT_TIMEOUT = 20
    JOIN_CHECK_THREAD_SECONDS = 1.0

    _channel: SyncAsyncQueueWriter
    running: Optional[bool]
    _iterate_sleep_seconds: Optional[float]
    _responsive_sleep_step_seconds: float
    pat_timeout: Optional[float]
    _last_pat_time: float

    def __init__(
        self,
        channel: Optional[SyncAsyncQueueWriter] = None,
        name: Optional[str] = None,
        iterate_sleep_seconds: Optional[float] = None,
        responsive_sleep_step_seconds: float = SLEEP_STEP_SECONDS,
        pat_timeout: Optional[float] = PAT_TIMEOUT,
        daemon: bool = True,
    ):
        super().__init__(name=name, daemon=daemon)
        if channel is None:
            self._channel = SyncAsyncQueueWriter()
        else:
            self._channel = channel
        self._iterate_sleep_seconds = iterate_sleep_seconds
        self._responsive_sleep_step_seconds = responsive_sleep_step_seconds
        self.running = None
        self.pat_timeout = pat_timeout
        self._last_pat_time = 0.0

    def _preiterate(self) -> None:
        pass

    def _iterate(self) -> None:
        pass

    def _handle_message(self, message: Any) -> None:
        pass

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def _handle_exception(self, exception: BaseException) -> bool:
        return False

    def request_stop(self) -> None:
        self.running = False

    def set_async_loop(
        self, loop: asyncio.AbstractEventLoop, async_queue: asyncio.Queue
    ) -> None:
        self._channel.set_async_loop(loop, async_queue)

    def set_async_loop_and_start(
        self, loop: asyncio.AbstractEventLoop, async_queue: asyncio.Queue
    ) -> None:
        self.set_async_loop(loop, async_queue)
        self.start()

    def put_to_sync_queue(
        self, message: Any, block: bool = True, timeout: Optional[float] = None
    ) -> None:
        self._channel.put_to_sync_queue(message, block=block, timeout=timeout)

    def _put_to_async_queue(self, message: Any) -> None:
        self._channel.put_to_async_queue(message)

    def run(self):
        if self.running is None:
            self.running = True
            self._last_pat_time = time.time()
            self._preiterate()
            while self.running:
                try:
                    if self._iterate_sleep_seconds is not None:
                        responsive_sleep(
                            self,
                            self._iterate_sleep_seconds,
                            running_field_name="running",
                            step_duration=self._responsive_sleep_step_seconds,
                        )
                    if self.running:
                        self._iterate()
                    if self.running and self.time_to_pat():
                        self.pat_watchdog()
                    if self.running and self._channel.sync_queue is not None:
                        try:
                            message = self._channel.get_from_sync_queue(block=False)
                            if self.running:
                                self._handle_message(message)
                        except queue.Empty:
                            pass
                except BaseException as e:
                    handle_exception_str = ""
                    try:
                        handled = self._handle_exception(e)
                    except BaseException as e2:
                        handled = False
                        handle_exception_str = traceback.format_exception(e2)
                    if not handled:
                        self.running = False
                        reason = (
                            f"SyncAsyncInteractionThread ({self.name}) caught exception:\n"
                            f"{traceback.format_exception(e)}\n"
                        )
                        if handle_exception_str:
                            reason += (
                                "While handling that exception, _handle_exception caused exception:\n"
                                f"{handle_exception_str}\n"
                            )
                        self._put_to_async_queue(
                            InternalShutdownMessage(Src=self.name, Reason=reason)
                        )

    def time_to_pat(self) -> bool:
        return time.time() >= (self._last_pat_time + (self.pat_timeout / 2))

    def pat_watchdog(self):
        self._last_pat_time = time.time()
        self._put_to_async_queue(PatInternalWatchdogMessage(src=self.name))

    async def async_join(self, timeout: float = None) -> None:
        await async_polling_thread_join(self, timeout)



import argparse
import asyncio
import logging
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Callable
from typing import Optional
from typing import Type
from typing import TypeVar

from gwproactor import Proactor
from gwproactor import ProactorSettings
from gwproactor import setup_logging
from gwproactor.config import DEFAULT_BASE_NAME
from gwproactor.config import LoggingSettings
from gwproactor.config import MQTTClient
from gwproactor.config import Paths
from gwproactor_test import copy_keys
from gwproactor_test.certs import uses_tls
from gwproactor_test.logger_guard import LoggerGuards
from gwproactor_test.proactor_recorder import ProactorT
from gwproactor_test.proactor_recorder import RecorderInterface
from gwproactor_test.proactor_recorder import make_recorder_class


@dataclass
class ProactorTestHelper:
    name: str
    path_name: str
    settings: ProactorSettings
    kwargs: dict = field(default_factory=dict)
    proactor: Optional[RecorderInterface] = None


ChildT = TypeVar("ChildT", bound=Proactor)
ParentT = TypeVar("ParentT", bound=Proactor)
ChildSettingsT = TypeVar("ChildSettingsT", bound=ProactorSettings)
ParentSettingsT = TypeVar("ParentSettingsT", bound=ProactorSettings)


class CommTestHelper:
    parent_t: Type[ProactorT]
    child_t: Type[Proactor]
    parent_settings_t: Type[ParentSettingsT]
    child_settings_t: Type[ProactorSettings]

    parent_recorder_t: Callable[..., RecorderInterface] = None
    child_recorder_t: Callable[..., RecorderInterface] = None

    parent_helper: ProactorTestHelper
    child_helper: ProactorTestHelper
    verbose: bool
    child_verbose: bool
    parent_verbose: bool
    parent_on_screen: bool
    lifecycle_logging: bool
    logger_guards: LoggerGuards

    warn_if_multi_subscription_tests_skipped: bool = True

    @classmethod
    def setup_class(cls):
        if cls.parent_recorder_t is None:
            cls.parent_recorder_t = make_recorder_class(cls.parent_t)
        if cls.child_recorder_t is None:
            cls.child_recorder_t = make_recorder_class(cls.child_t)

    def __init__(
        self,
        child_settings: Optional[ChildSettingsT] = None,
        parent_settings: Optional[ParentSettingsT] = None,
        verbose: bool = False,
        child_verbose: bool = False,
        parent_verbose: bool = False,
        lifecycle_logging: bool = False,
        add_child: bool = False,
        add_parent: bool = False,
        start_child: bool = False,
        start_parent: bool = False,
        parent_on_screen: bool = False,
        child_name: str = "",
        parent_name: str = "",
        child_path_name: str = "child",
        parent_path_name: str = "parent",
        child_kwargs: Optional[dict] = None,
        parent_kwargs: Optional[dict] = None,
    ):
        self.setup_class()
        self.child_helper = ProactorTestHelper(
            child_name,
            child_path_name,
            (
                self.child_settings_t(paths=Paths(name=Path(child_path_name)))
                if child_settings is None
                else child_settings
            ),
            dict() if child_kwargs is None else child_kwargs,
        )
        self.parent_helper = ProactorTestHelper(
            parent_name,
            parent_path_name,
            (
                self.parent_settings_t(
                    logging=LoggingSettings(
                        base_log_name=f"parent_{DEFAULT_BASE_NAME}"
                    ),
                    paths=Paths(name=Path(parent_path_name)),
                )
                if parent_settings is None
                else parent_settings
            ),
            dict() if parent_kwargs is None else parent_kwargs,
        )
        self.verbose = verbose
        self.child_verbose = child_verbose
        self.parent_verbose = parent_verbose
        self.parent_on_screen = parent_on_screen
        self.lifecycle_logging = lifecycle_logging
        self.setup_logging()
        if add_child or start_child:
            self.add_child()
            if start_child:
                self.start_child()
        if add_parent or start_parent:
            self.add_parent()
            if start_parent:
                self.start_parent()

    @classmethod
    def _make(
        cls, recorder_t: Callable[..., RecorderInterface], helper: ProactorTestHelper
    ) -> RecorderInterface:
        if uses_tls(helper.settings):
            copy_keys(helper.path_name, helper.settings)
        return recorder_t(helper.name, helper.settings, **helper.kwargs)

    def make_parent(self) -> RecorderInterface:
        return self._make(self.parent_recorder_t, self.parent_helper)

    def make_child(self) -> RecorderInterface:
        return self._make(self.child_recorder_t, self.child_helper)

    @property
    def parent(self) -> Optional[ProactorT]:
        return self.parent_helper.proactor

    @property
    def child(self) -> Optional[ProactorT]:
        return self.child_helper.proactor

    def start_child(self) -> "CommTestHelper":
        if self.child is not None:
            self.start_proactor(self.child)
        return self

    def start_parent(self) -> "CommTestHelper":
        if self.parent is not None:
            self.start_proactor(self.parent)
        return self

    def start_proactor(self, proactor: Proactor) -> "CommTestHelper":
        asyncio.create_task(proactor.run_forever(), name=f"{proactor.name}_run_forever")
        return self

    def start(self) -> "CommTestHelper":
        return self

    def add_child(self) -> "CommTestHelper":
        self.child_helper.proactor = self.make_child()
        return self

    def add_parent(self) -> "CommTestHelper":
        self.parent_helper.proactor = self.make_parent()
        return self

    def remove_child(self) -> "CommTestHelper":
        self.child_helper.proactor = None
        return self

    def remove_parent(self) -> "CommTestHelper":
        self.parent_helper.proactor = None
        return self

    @classmethod
    def _get_clients_supporting_tls(
        cls, settings: ProactorSettings
    ) -> list[MQTTClient]:
        clients = []
        for _, v in settings._iter():  # noqa
            if isinstance(v, MQTTClient):
                clients.append(v)
        return clients

    def _get_child_clients_supporting_tls(self) -> list[MQTTClient]:
        """Overide to filter which MQTT clients of ChildSettingsT are treated as supporting TLS"""
        return self._get_clients_supporting_tls(self.child_helper.settings)

    def _get_parent_clients_supporting_tls(self) -> list[MQTTClient]:
        """Overide to filter which MQTT clients of ParentSettingsT are treated as supporting TLS"""
        return self._get_clients_supporting_tls(self.parent_helper.settings)

    @classmethod
    def _set_settings_use_tls(cls, use_tls: bool, clients: list[MQTTClient]) -> None:
        for client in clients:
            client.tls.use_tls = use_tls

    def set_use_tls(self, use_tls: bool) -> None:
        """Set MQTTClients which support TLS in parent and child settings to use TLS per use_tls. Clients supporting TLS
        is determined by _get_child_clients_supporting_tls() and _get_parent_clients_supporting_tls() which may be
        overriden in derived class.
        """
        self._set_settings_use_tls(use_tls, self._get_child_clients_supporting_tls())
        self._set_settings_use_tls(use_tls, self._get_parent_clients_supporting_tls())

    def setup_logging(self):
        self.child_helper.settings.paths.mkdirs(parents=True)
        self.parent_helper.settings.paths.mkdirs(parents=True)
        errors = []
        if not self.lifecycle_logging:
            self.child_helper.settings.logging.levels.lifecycle = logging.WARNING
            self.parent_helper.settings.logging.levels.lifecycle = logging.WARNING
        self.logger_guards = LoggerGuards(
            list(self.child_helper.settings.logging.qualified_logger_names().values())
            + list(
                self.parent_helper.settings.logging.qualified_logger_names().values()
            )
        )
        setup_logging(
            argparse.Namespace(verbose=self.verbose or self.child_verbose),
            self.child_helper.settings,
            errors,
            add_screen_handler=True,
            root_gets_handlers=False,
        )
        assert not errors
        setup_logging(
            argparse.Namespace(verbose=self.verbose or self.parent_verbose),
            self.parent_helper.settings,
            errors,
            add_screen_handler=self.parent_on_screen,
            root_gets_handlers=False,
        )
        assert not errors

    async def stop_and_join(self):
        proactors = [
            helper.proactor
            for helper in [self.child_helper, self.parent_helper]
            if helper.proactor is not None
        ]
        for proactor in proactors:
            # noinspection PyBroadException
            try:
                proactor.stop()
            except:
                pass
        for proactor in proactors:
            # noinspection PyBroadException
            try:
                await proactor.join()
            except:
                pass

    async def __aenter__(self) -> "CommTestHelper":
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.stop_and_join()
        if exc is not None:
            logging.getLogger("gridworks").error(
                "CommTestHelper caught error %s.\nWorking log dirs:\n\t[%s]\n\t[%s]",
                exc,
                self.child_helper.settings.paths.log_dir,
                self.parent_helper.settings.paths.log_dir,
            )
        self.logger_guards.restore()

    def summary_str(self) -> str:
        return (
            "CHILD:\n"
            f"{self.child.summary_str()}\n"
            "PARENT:\n"
            f"{self.parent.summary_str()}"
        )



"""Code for actors that use a simple rest interaction, converting the response to one or more
REST commands into a message posted to main processing thread.

"""

import asyncio
from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Optional

import aiohttp as aiohttp
import yarl
from aiohttp import ClientResponse
from aiohttp import ClientSession
from aiohttp import ClientTimeout
from gwproto import Message
from gwproto.data_classes.components.rest_poller_component import RESTPollerComponent
from gwproto.type_helpers import AioHttpClientTimeout
from gwproto.type_helpers import RESTPollerSettings
from gwproto.type_helpers import URLConfig
from result import Result

from gwproactor import Actor
from gwproactor import Problems
from gwproactor import ServicesInterface
from gwproactor.proactor_interface import INVALID_IO_TASK_HANDLE
from gwproactor.proactor_interface import IOLoopInterface


Converter = Callable[[ClientResponse], Awaitable[Optional[Message]]]
ThreadSafeForwarder = Callable[[Message], Any]


async def null_converter(response: ClientResponse) -> Optional[Message]:  # noqa
    return None


def null_forwarder(message: Message) -> None:  # noqa
    return None


def to_client_timeout(
    timeout: Optional[AioHttpClientTimeout],
) -> Optional[ClientTimeout]:
    if timeout is not None:
        return ClientTimeout(**timeout.dict())
    return None


@dataclass
class SessionArgs:
    base_url: Optional[yarl.URL] = None
    kwargs: dict = field(default_factory=dict)


@dataclass
class RequestArgs:
    method: str = "GET"
    url: Optional[yarl.URL] = None
    kwargs: dict = field(default_factory=dict)


class RESTPoller:
    _name: str
    _rest: RESTPollerSettings
    _io_loop_manager: IOLoopInterface
    _task_id: int
    _session_args: Optional[SessionArgs] = None
    _request_args: Optional[RequestArgs] = None
    _converter: Converter
    _forward: ThreadSafeForwarder

    def __init__(
        self,
        name: str,
        rest: RESTPollerSettings,
        loop_manager: IOLoopInterface,
        convert: Converter = null_converter,
        forward: ThreadSafeForwarder = null_forwarder,
        cache_request_args: bool = True,
    ):
        self._name = name
        self._task_id = INVALID_IO_TASK_HANDLE
        self._rest = rest
        self._io_loop_manager = loop_manager
        self._convert = convert
        self._forward = forward
        if cache_request_args:
            self._session_args = self._make_session_args()
            self._request_args = self._make_request_args()

    def _make_base_url(self) -> Optional[yarl.URL]:
        return URLConfig.make_url(self._rest.session.base_url)

    def _make_url(self) -> Optional[yarl.URL]:
        return URLConfig.make_url(self._rest.request.url)

    def _make_session_args(self) -> SessionArgs:
        return SessionArgs(
            self._make_base_url(),
            dict(
                self._rest.session.dict(
                    exclude={"base_url", "timeout"},
                    exclude_unset=True,
                ),
                timeout=to_client_timeout(self._rest.session.timeout),
            ),
        )

    def _make_request_args(self) -> RequestArgs:
        return RequestArgs(
            self._rest.request.method,
            self._make_url(),
            dict(
                self._rest.request.dict(
                    exclude={"method", "url", "timeout"},
                    exclude_unset=True,
                ),
                timeout=to_client_timeout(self._rest.request.timeout),
            ),
        )

    async def _make_request(self, session: ClientSession) -> Optional[ClientResponse]:
        try:
            args = self._request_args
            if args is None:
                args = self._make_request_args()
            response = await session.request(args.method, args.url, **args.kwargs)
            if self._rest.errors.request.error_for_http_status:
                response.raise_for_status()
        except BaseException as e:
            response = None
            if self._rest.errors.request.report:
                try:
                    self._forward(
                        Message(
                            Payload=Problems(errors=[e]).problem_event(
                                summary=(
                                    f"Request error for <{self._name}>: {type(e)} <{e}>"
                                ),
                            )
                        )
                    )
                except:  # noqa
                    pass
            if self._rest.errors.request.raise_exception:
                raise e
        return response

    async def _convert(self, response: ClientResponse) -> Optional[Message]:
        try:
            message = await self._converter(response)
        except BaseException as convert_exception:
            message = None
            if self._rest.errors.convert.report:
                try:
                    self._forward(
                        Message(
                            Payload=Problems(errors=[convert_exception]).problem_event(
                                summary=(
                                    f"Convert error for <{self._name}>: {type(convert_exception)} <{convert_exception}>"
                                ),
                            )
                        )
                    )
                except:  # noqa
                    pass
            if self._rest.errors.convert.raise_exception:
                raise convert_exception
        return message

    def _get_next_sleep_seconds(self) -> float:
        return self._rest.poll_period_seconds

    async def _run(self):
        try:
            reconnect = True
            while reconnect:
                args = self._session_args
                if args is None:
                    args = self._make_session_args()
                async with aiohttp.ClientSession(
                    args.base_url, **args.kwargs
                ) as session:
                    while True:
                        response = await self._make_request(session)
                        if response is not None:
                            async with response:
                                message = await self._convert(response)
                            if message is not None:
                                self._forward(message)
                        sleep_seconds = self._get_next_sleep_seconds()
                        await asyncio.sleep(sleep_seconds)

        except BaseException as e:
            raise e

    def start(self) -> None:
        self._task_id = self._io_loop_manager.add_io_coroutine(
            self._run(),
            name=self._name,
        )

    def stop(self) -> None:
        self._io_loop_manager.cancel_io_routine(self._task_id)


class RESTPollerActor(Actor):
    _poller: RESTPoller

    def __init__(
        self,
        name: str,
        services: ServicesInterface,
        convert: Converter = null_converter,
        forward: ThreadSafeForwarder = null_forwarder,
        cache_request_args: bool = True,
    ):
        super().__init__(name, services)
        component = services.hardware_layout.component(self.name)
        if not isinstance(component, RESTPollerComponent):
            display_name = getattr(
                component, "display_name", "MISSING ATTRIBUTE display_name"
            )
            raise ValueError(
                f"ERROR. Component <{display_name}> has type {type(component)}. "
                f"Expected RESTPollerComponent.\n"
                f"  Node: {self.name}\n"
                f"  Component id: {component.component_id}"
            )
        self._poller = RESTPoller(
            name,
            component.rest,
            services.io_loop_manager,
            convert=convert,
            forward=forward,
            cache_request_args=cache_request_args,
        )

    def process_message(self, message: Message) -> Result[bool, BaseException]:
        pass

    def start(self) -> None:
        self._poller.start()

    def stop(self) -> None:
        self._poller.stop()

    async def join(self) -> None:
        """IOLoop will take care of shutting down the associated task."""



"""Local pytest configuration"""

import contextlib
import os
import shutil
from pathlib import Path
from types import NoneType
from typing import Generator
from typing import Optional

import dotenv
import pytest
from _pytest.monkeypatch import MonkeyPatch

from gwproactor import ProactorSettings
from gwproactor.config import DEFAULT_LAYOUT_FILE
from gwproactor.config import Paths
from gwproactor_test.dummies import DUMMY_CHILD_ENV_PREFIX
from gwproactor_test.dummies import DUMMY_PARENT_ENV_PREFIX


TEST_DOTENV_PATH = "tests/.env-gwproactor-test"
TEST_DOTENV_PATH_VAR = "GWPROACTOR_TEST_DOTENV_PATH"
TEST_HARDWARE_LAYOUT_PATH = Path(__file__).parent / "config" / DEFAULT_LAYOUT_FILE
DUMMY_TEST_HARDWARE_LAYOUT_PATH = (
    Path(__file__).parent / "config" / "dummy-hardware-layout.json"
)


class DefaultTestEnv:
    """Context manager for monkeypatched environment with:
        - all vars starting with any entry in DEFAULT_PREFIXES removed
        - vars loaded from test env file, if specified
        - xdg vars set relative to passed in xdg_home parameter
        - working config directory created via xdg_home
        - test hardware layout file copied into working config directory.

    >>> tmp_path = Path("/home/bla")
    >>> with DefaultTestEnv(tmp_path).context() as mpatch:
    ...     assert ProactorSettings().paths.hardware_layout == Path("/home/bla/.config/gridworks/scada/hardware-layout.json")
    ...     assert ProactorSettings().paths.hardware_layout.exists()


    The default test env file is tests/.env-gwproactor-test. This path can be overridden with the environment variable
    GWPROACTOR_TEST_DOTENV_PATH. The test env file will be ignored if the GWPROACTOR_TEST_DOTENV_PATH environment
    variable exists but is empty or the specified path does not exist.

    Hardware file copying can be suppressed by passing copy_test_layout as False.

    Working test directory creation can be suppressed by passing xdg_home as None.
    """

    DEFAULT_PREFIXES = [
        DUMMY_CHILD_ENV_PREFIX,
        DUMMY_PARENT_ENV_PREFIX,
    ]

    xdg_home: Path | NoneType = None
    src_test_layout: Path = TEST_HARDWARE_LAYOUT_PATH
    copy_test_layout: bool = True
    use_test_dotenv: bool = True
    prefixes: Optional[list[str]] = None

    def __init__(
        self,
        xdg_home: Path | str | None = None,
        src_test_layout: Path = TEST_HARDWARE_LAYOUT_PATH,
        copy_test_layout: bool = True,
        use_test_dotenv: bool = True,
        prefixes: Optional[list[str]] = None,
    ):
        if isinstance(xdg_home, str) and bool(xdg_home):
            xdg_home = Path(xdg_home)
        self.xdg_home = xdg_home
        self.src_test_layout = src_test_layout
        self.copy_test_layout = copy_test_layout
        self.use_test_dotenv = use_test_dotenv
        if prefixes is None:
            self.prefixes = self.get_default_prefixes()
        else:
            self.prefixes = prefixes[:]

    @classmethod
    def get_default_prefixes(cls):
        return cls.DEFAULT_PREFIXES[:]

    @contextlib.contextmanager
    def context(self) -> Generator[MonkeyPatch, None, None]:
        """Produce monkeypatch context manager from this object"""
        mpatch = MonkeyPatch()
        with mpatch.context() as m:
            self.clean_env(m)
            self.load_test_dotenv()
            self.setup_text_xdg_home(m)
            yield m

    def setup_text_xdg_home(self, m: MonkeyPatch):
        if self.xdg_home is not None:
            m.setenv("XDG_DATA_HOME", str(self.xdg_home / ".local" / "share"))
            m.setenv("XDG_STATE_HOME", str(self.xdg_home / ".local" / "state"))
            m.setenv("XDG_CONFIG_HOME", str(self.xdg_home / ".config"))
            if self.copy_test_layout:
                paths = Paths()
                paths.hardware_layout.parent.mkdir(parents=True, exist_ok=True)
                shutil.copyfile(self.src_test_layout, paths.hardware_layout)

    def clean_env(self, m: MonkeyPatch):
        for env_var in os.environ:
            for prefix in self.prefixes:
                if env_var.startswith(prefix):
                    m.delenv(env_var)

    def load_test_dotenv(self):
        if self.use_test_dotenv:
            test_dotenv_file = os.getenv(TEST_DOTENV_PATH_VAR)
            if test_dotenv_file is None:
                test_dotenv_file = TEST_DOTENV_PATH
            if test_dotenv_file:
                test_dotenv_path = Path(test_dotenv_file)
                if test_dotenv_path.exists():
                    dotenv.load_dotenv(dotenv_path=test_dotenv_path)


@pytest.fixture(autouse=True)
def default_test_env(request, tmp_path) -> Generator[MonkeyPatch, None, None]:
    """Automatically used fixture producing monkeypatched environment with:
        - all vars starting with any entry in DefaultTestEnv.DEFAULT_PREFIXES removed
        - vars loaded from test env file, if specified
        - xdg vars set relative to passed in xdg_home parameter
        - working config directory created via xdg_home
        - test hardware layout file copied into working config directory.

    Note that this fixture is run before _every_ test.

    The behavior of this fixture can be customized by:
        1. Modifying the contents of tests/.env-gwproactor-test.
        2. Changing the the path to the test dotenv file via the GWPROACTOR_TEST_DOTENV_PATH environment variable.
        3. Explicitly passing and parametrizing this fixture. For example, to run a test with a different hardware
            layout file, such as DUMMY_TEST_HARDWARE_LAYOUT_PATH:

        >>> from gwproactor_test.clean import DUMMY_TEST_HARDWARE_LAYOUT_PATH
        >>> @pytest.mark.parametrize("default_test_env", [(DefaultTestEnv(src_test_layout=DUMMY_TEST_HARDWARE_LAYOUT_PATH)], indirect=True)
        >>> def test_something(default_test_env):
        >>>    assert Paths().hardware_layout.open().read() == DUMMY_TEST_HARDWARE_LAYOUT_PATH.open().read()


    """
    test_env = getattr(request, "param", DefaultTestEnv())
    if test_env.xdg_home is None:
        test_env.xdg_home = tmp_path
    with test_env.context() as mpatch:
        yield mpatch


@pytest.fixture
def clean_test_env(request, tmp_path) -> Generator[MonkeyPatch, None, None]:
    """Get a monkeypatched environment with all vars starting with any entry in DEFAULT_PREFIXES *removed* (and none
    loaded from any dotenv file)."""
    test_env = getattr(request, "param", DefaultTestEnv(use_test_dotenv=False))
    if test_env.xdg_home is None:
        test_env.xdg_home = tmp_path
    with test_env.context() as mpatch:
        yield mpatch



import asyncio
import logging
import time
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Type

import pytest
from gwproto import MQTTTopic
from paho.mqtt.client import MQTT_ERR_CONN_LOST

from gwproactor import ServicesInterface
from gwproactor.links import StateName
from gwproactor.message import DBGEvent
from gwproactor.message import DBGPayload
from gwproactor.persister import TimedRollingFilePersister
from gwproactor_test.certs import uses_tls
from gwproactor_test.comm_test_helper import CommTestHelper
from gwproactor_test.wait import await_for


@dataclass
class _EventEntry:
    uid: str
    path: Path


class _EventGen:
    ok: list[_EventEntry]
    corrupt: list[_EventEntry]
    empty: list[_EventEntry]
    missing: list[_EventEntry]

    persister: TimedRollingFilePersister

    def __len__(self) -> int:
        return len(self.ok) + len(self.corrupt) + len(self.empty)

    def __init__(self, proactor: ServicesInterface):
        self.ok = []
        self.corrupt = []
        self.empty = []
        self.missing = []
        persister = proactor._event_persister  # noqa
        assert isinstance(persister, TimedRollingFilePersister)
        self.persister = persister

    def _generate_event(self, member_name: str) -> _EventEntry:
        event = DBGEvent(Command=DBGPayload(), Msg=f"event {len(self)} {member_name}")
        ret = self.persister.persist(
            event.MessageId, event.json(sort_keys=True, indent=2).encode()
        )
        if ret.is_err():
            raise ret.err()
        entry = _EventEntry(
            event.MessageId, self.persister.get_path(event.MessageId)  # noqa
        )
        getattr(self, member_name).append(entry)
        return entry

    def _generate_ok(self) -> _EventEntry:
        return self._generate_event("ok")

    def _generate_corrupt(self) -> _EventEntry:
        entry = self._generate_event("corrupt")
        with entry.path.open() as f:
            contents = f.read()
        with entry.path.open("w") as f:
            f.write(contents[:-6])
        return entry

    def _generate_empty(self) -> _EventEntry:
        entry = self._generate_event("empty")
        with entry.path.open("w") as f:
            f.write("")
        return entry

    def _generate_missing(self) -> _EventEntry:
        entry = self._generate_event("missing")
        entry.path.unlink()
        return entry

    def generate(
        self,
        num_ok: int = 0,
        num_corrupt: int = 0,
        num_empty: int = 0,
        num_missing: int = 0,
    ):
        for _ in range(num_ok):
            self._generate_ok()
        for _ in range(num_corrupt):
            self._generate_corrupt()
        for _ in range(num_empty):
            self._generate_empty()
        for _ in range(num_missing):
            self._generate_missing()


@pytest.mark.asyncio
class ProactorCommTests:
    CTH: Type[CommTestHelper]

    async def test_no_parent(self):
        async with self.CTH(add_child=True) as h:
            child = h.child
            stats = child.stats.link(child.upstream_client)
            comm_event_counts = stats.comm_event_counts
            link = child._links.link(child.upstream_client)

            # unstarted child
            assert stats.num_received == 0
            assert link.state == StateName.not_started
            child.logger.info(child.settings.json(sort_keys=True, indent=2))

            # start child
            h.start_child()
            await await_for(
                link.active_for_send,
                1,
                "ERROR waiting link active_for_send",
                err_str_f=child.summary_str,
            )
            assert not link.active_for_recv()
            assert not link.active()
            assert link.state == StateName.awaiting_peer
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert len(stats.comm_events) == 2
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child._event_persister

            # Tell client we lost comm.
            child.mqtt_client_wrapper("gridworks")._client._loop_rc_handle(
                MQTT_ERR_CONN_LOST
            )

            # Wait for reconnect
            await await_for(
                lambda: stats.comm_event_counts[
                    "gridworks.event.comm.mqtt.fully.subscribed"
                ]
                > 1,
                3,
                "ERROR waiting link to resubscribe after comm loss",
                err_str_f=child.summary_str,
            )
            assert link.active_for_send()
            assert not link.active_for_recv()
            assert not link.active()
            assert link.state == StateName.awaiting_peer
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 2
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 2
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 1
            assert len(stats.comm_events) == 5
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child._event_persister

    async def test_basic_comm_child_first(self):
        async with self.CTH(add_child=True, add_parent=True) as h:
            child = h.child
            child_stats = child.stats.link(child.upstream_client)
            child_comm_event_counts = child_stats.comm_event_counts
            child_link = child._links.link(child.upstream_client)

            # unstarted child, parent
            assert child_stats.num_received == 0
            assert child_link.state == StateName.not_started

            # start child
            h.start_child()
            await await_for(
                child_link.active_for_send,
                1,
                "ERROR waiting link active_for_send",
                err_str_f=child.summary_str,
            )
            assert not child_link.active_for_recv()
            assert not child_link.active()
            assert child_link.state == StateName.awaiting_peer
            assert child_comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert (
                child_comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"]
                == 1
            )
            assert child_comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert child_comm_event_counts["gridworks.event.comm.peer.active"] == 0
            assert len(child_stats.comm_events) == 2
            for comm_event in child_stats.comm_events:
                assert comm_event.MessageId in child._event_persister

            # start parent
            h.start_parent()

            # wait for link to go active
            await await_for(
                child_link.active,
                10,
                "ERROR waiting link active",
                err_str_f=child.summary_str,
            )
            assert child_link.active_for_recv()
            assert child_link.active()
            assert child_link.state == StateName.active
            assert child_comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert (
                child_comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"]
                == 1
            )
            assert child_comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert child_comm_event_counts["gridworks.event.comm.peer.active"] == 1
            assert len(child_stats.comm_events) == 3

            # wait for all events to be acked
            await await_for(
                lambda: child._event_persister.num_pending == 0,
                1,
                "ERROR waiting for events to be acked",
                err_str_f=child.summary_str,
            )

            # Tell client we lost comm.
            child.mqtt_client_wrapper("gridworks")._client._loop_rc_handle(
                MQTT_ERR_CONN_LOST
            )

            # Wait for reconnect
            await await_for(
                lambda: child_stats.comm_event_counts[
                    "gridworks.event.comm.peer.active"
                ]
                > 1,
                3,
                "ERROR waiting link to resubscribe after comm loss",
                err_str_f=child.summary_str,
            )
            assert child_link.active_for_send()
            assert child_link.active_for_recv()
            assert child_link.active()
            assert child_link.state == StateName.active
            assert child_comm_event_counts["gridworks.event.comm.mqtt.connect"] == 2
            assert (
                child_comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"]
                == 2
            )
            assert child_comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 1
            assert child_comm_event_counts["gridworks.event.comm.peer.active"] == 2
            assert len(child_stats.comm_events) == 7

            # wait for all events to be acked
            await await_for(
                lambda: child._event_persister.num_pending == 0,
                1,
                "ERROR waiting for events to be acked",
                err_str_f=child.summary_str,
            )

    # @pytest.mark.asyncio
    # async def test_broker_dns_failure(self):
    #     """Verify proactor does not crash in presence of bad host.
    #
    #     This test commented out because it's very slow - 4 seconds before mqtt client thread
    #     reports first problem.
    #
    #     A better test would:
    #         - be faster
    #         - verify client thread continues
    #         - verify problem reports occur
    #         - possibly verify that eventually child requests device restart since pi's are
    #           having DNS failures on network comm loss that seem to be undone by pi restart
    #     """
    #     no_broker = MQTTClient(host="www.foo.com")
    #     async with self.CTH(
    #         child_settings=DummyChildSettings(
    #             parent_mqtt=no_broker,
    #         ),
    #         start_child=True,
    #         verbose=True,
    #     ):
    #         for i in range(20):
    #             print(f"{i}...")
    #             await asyncio.sleep(1)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("suppress_tls", [False, True])
    async def test_basic_comm_parent_first(self, request, suppress_tls: bool):
        async with self.CTH() as h:
            base_logger = logging.getLogger(
                h.child_helper.settings.logging.base_log_name
            )
            base_logger.warning(f"{request.node.name}  suppress_tls: {suppress_tls}")
            if suppress_tls:
                if not uses_tls(h.child_helper.settings) and not uses_tls(
                    h.parent_helper.settings
                ):
                    base_logger.warning(
                        "Skipping test <%s> since TLS has already been suppressed by environment variables",
                        request.node.name,
                    )
                else:
                    h.set_use_tls(False)
            h.add_child()
            h.add_parent()
            child = h.child
            child_stats = child.stats.link(child.upstream_client)
            child_comm_event_counts = child_stats.comm_event_counts
            child_link = child._links.link(child.upstream_client)
            parent = h.parent
            # TODO: hack
            parent_link = parent._links.link(parent.primary_peer_client)

            # unstarted parent
            assert parent_link.state == StateName.not_started

            # start parent
            h.start_parent()
            await await_for(
                parent_link.active_for_send,
                1,
                "ERROR waiting link active_for_send",
                err_str_f=parent.summary_str,
            )

            # unstarted child
            assert child_stats.num_received == 0
            assert child_link.state == StateName.not_started

            # start child
            h.start_child()
            await await_for(
                child_link.active,
                1,
                "ERROR waiting link active",
                err_str_f=parent.summary_str,
            )

            assert child_link.active_for_recv()
            assert child_link.active()
            assert child_link.state == StateName.active
            assert child_comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert (
                child_comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"]
                == 1
            )
            assert child_comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert child_comm_event_counts["gridworks.event.comm.peer.active"] == 1
            assert len(child_stats.comm_events) == 3

            # wait for all events to be acked
            await await_for(
                lambda: child._event_persister.num_pending == 0,
                1,
                "ERROR waiting for events to be acked",
                err_str_f=child.summary_str,
            )

    @pytest.mark.asyncio
    async def test_basic_parent_comm_loss(self):
        async with self.CTH(add_child=True, add_parent=True, verbose=False) as h:
            child = h.child
            child_stats = child.stats.link(child.upstream_client)
            child_comm_event_counts = child_stats.comm_event_counts
            child_link = child._links.link(child.upstream_client)
            parent = h.parent
            parent_link = parent._links.link(parent.primary_peer_client)

            # unstarted child, parent
            assert parent_link.state == StateName.not_started
            assert child_stats.num_received == 0
            assert child_link.state == StateName.not_started

            # start child, parent
            h.start_child()
            h.start_parent()
            await await_for(
                child_link.active,
                1,
                "ERROR waiting link active",
                err_str_f=child.summary_str,
            )
            assert child_link.active_for_recv()
            assert child_link.active()
            assert child_link.state == StateName.active
            assert child_comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert (
                child_comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"]
                == 1
            )
            assert child_comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert child_comm_event_counts["gridworks.event.comm.peer.active"] == 1
            assert len(child_stats.comm_events) == 3

            # wait for all events to be acked
            await await_for(
                lambda: child._event_persister.num_pending == 0,
                1,
                "ERROR waiting for events to be acked",
                err_str_f=child.summary_str,
            )

            # Tell *child* client we lost comm.
            child.mqtt_client_wrapper(child.upstream_client)._client._loop_rc_handle(
                MQTT_ERR_CONN_LOST
            )

            # Wait for reconnect
            await await_for(
                lambda: child_stats.comm_event_counts[
                    "gridworks.event.comm.peer.active"
                ]
                > 1,
                3,
                "ERROR waiting link to resubscribe after comm loss",
                err_str_f=child.summary_str,
            )
            assert child_link.active_for_send()
            assert child_link.active_for_recv()
            assert child_link.active()
            assert child_link.state == StateName.active
            assert child_comm_event_counts["gridworks.event.comm.mqtt.connect"] == 2
            assert (
                child_comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"]
                == 2
            )
            assert child_comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 1
            assert child_comm_event_counts["gridworks.event.comm.peer.active"] == 2
            assert len(child_stats.comm_events) == 7

            # wait for all events to be acked
            await await_for(
                lambda: child._event_persister.num_pending == 0,
                1,
                "ERROR waiting for events to be acked",
                err_str_f=child.summary_str,
            )

            # Tell *parent* client we lost comm.
            parent.mqtt_client_wrapper(
                parent.primary_peer_client
            )._client._loop_rc_handle(MQTT_ERR_CONN_LOST)
            # wait for child to get ping from parent when parent reconnects to mqtt
            parent_ping_topic = MQTTTopic.encode(
                "gw", parent.publication_name, "gridworks-ping"
            )
            num_parent_pings = child_stats.num_received_by_topic[parent_ping_topic]
            await await_for(
                lambda: child_stats.num_received_by_topic[parent_ping_topic]
                == num_parent_pings + 1,
                3,
                f"ERROR waiting for parent ping {parent_ping_topic}",
                err_str_f=child.summary_str,
            )
            # verify no child comm state change has occurred.
            err_str = f"\n{child.summary_str()}\n" f"{parent.summary_str()}\n"
            assert child_link.active_for_send()
            assert child_link.active_for_recv()
            assert child_link.active()
            assert child_link.state == StateName.active, err_str
            assert (
                child_comm_event_counts["gridworks.event.comm.mqtt.connect"] == 2
            ), err_str
            assert (
                child_comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"]
                == 2
            ), err_str
            assert (
                child_comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 1
            ), err_str
            assert (
                child_comm_event_counts["gridworks.event.comm.peer.active"] == 2
            ), err_str
            assert len(child_stats.comm_events) == 7, err_str
            assert child._event_persister.num_pending == 0, err_str

            # Tell *both* clients we lost comm.
            parent.mqtt_client_wrapper(
                parent.primary_peer_client
            )._client._loop_rc_handle(MQTT_ERR_CONN_LOST)
            child.mqtt_client_wrapper(child.upstream_client)._client._loop_rc_handle(
                MQTT_ERR_CONN_LOST
            )

            # Wait for reconnect
            await await_for(
                lambda: child_stats.comm_event_counts[
                    "gridworks.event.comm.peer.active"
                ]
                > 2,
                3,
                "ERROR waiting link to resubscribe after comm loss",
                err_str_f=child.summary_str,
            )
            assert child_link.active_for_send()
            assert child_link.active_for_recv()
            assert child_link.active()
            assert child_link.state == StateName.active
            assert child_comm_event_counts["gridworks.event.comm.mqtt.connect"] == 3
            assert (
                child_comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"]
                == 3
            )
            assert child_comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 2
            assert child_comm_event_counts["gridworks.event.comm.peer.active"] == 3
            assert len(child_stats.comm_events) == 11

            # wait for all events to be acked
            await await_for(
                lambda: child._event_persister.num_pending == 0,
                1,
                "ERROR waiting for events to be acked",
                err_str_f=child.summary_str,
            )

    @pytest.mark.asyncio
    async def test_awaiting_setup_and_peer(self):
        """
        Test:
         (connecting -> connected -> awaiting_setup_and_peer)
         (awaiting_setup_and_peer -> mqtt_suback -> awaiting_peer)
         (awaiting_setup_and_peer -> disconnected -> connecting)
        """
        async with self.CTH(add_child=True) as h:
            child = h.child
            stats = child.stats.link(child.upstream_client)
            comm_event_counts = stats.comm_event_counts
            link = child._links.link(child.upstream_client)

            # unstarted child
            assert stats.num_received == 0
            assert link.state == StateName.not_started

            # start child
            child.pause_subacks()
            h.start_child()
            await await_for(
                lambda: len(child.pending_subacks) == 1,
                1,
                "ERROR waiting suback pending",
                err_str_f=child.summary_str,
            )
            assert not link.active_for_send()
            assert not link.active_for_recv()
            assert not link.active()
            assert link.state == StateName.awaiting_setup_and_peer
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 0
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert len(stats.comm_events) == 1
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child._event_persister

            # Allow suback to arrive
            child.release_subacks()
            await await_for(
                lambda: link.in_state(StateName.awaiting_peer),
                1,
                f"ERROR waiting mqtt_suback",
                err_str_f=child.summary_str,
            )
            assert link.active_for_send()
            assert not link.active_for_recv()
            assert not link.active()
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert len(stats.comm_events) == 2
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child._event_persister

            # Tell client we lost comm
            child.pause_subacks()
            child.mqtt_client_wrapper(child.upstream_client)._client._loop_rc_handle(
                MQTT_ERR_CONN_LOST
            )
            await await_for(
                lambda: len(child.pending_subacks) == 1,
                3,
                "ERROR waiting suback pending",
                err_str_f=child.summary_str,
            )
            assert not link.active_for_send()
            assert not link.active_for_recv()
            assert not link.active()
            assert link.state == StateName.awaiting_setup_and_peer
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 2
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 1
            assert len(stats.comm_events) == 4
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child._event_persister

            # Tell client we lost comm
            child.pending_subacks = []
            child.mqtt_client_wrapper(child.upstream_client)._client._loop_rc_handle(
                MQTT_ERR_CONN_LOST
            )
            await await_for(
                lambda: len(stats.comm_events) > 4,
                1,
                "ERROR waiting comm fail",
                err_str_f=child.summary_str,
            )
            await await_for(
                lambda: link.in_state(StateName.awaiting_setup_and_peer),
                3,
                "ERROR waiting comm restore",
                err_str_f=child.summary_str,
            )
            assert not link.active_for_send()
            assert not link.active_for_recv()
            assert not link.active()
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 3
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 2
            assert len(stats.comm_events) == 6
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child._event_persister

            # Allow suback to arrive
            child.release_subacks()
            await await_for(
                lambda: link.in_state(StateName.awaiting_peer),
                1,
                f"ERROR waiting mqtt_suback",
                err_str_f=child.summary_str,
            )
            assert link.active_for_send()
            assert not link.active_for_recv()
            assert not link.active()
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 3
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 2
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 2
            assert len(stats.comm_events) == 7
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child._event_persister

    @pytest.mark.asyncio
    async def test_awaiting_setup_and_peer_corner_cases(self, request):
        """
        Test corner cases:
         (connecting -> connected -> awaiting_setup_and_peer)
         (awaiting_setup_and_peer -> mqtt_suback -> awaiting_setup_and_peer)
         (awaiting_setup_and_peer -> mqtt_suback -> awaiting_peer)
         (awaiting_setup_and_peer -> message_from_peer -> awaiting_setup)
        Force 1 suback per subscription. By default MQTTClientWrapper packs as many subscriptions as possible into a
        single subscribe message, so by default child only receives a single suback for all subscriptions.
        So that we can test (awaiting_setup_and_peer -> mqtt_suback -> awaiting_setup_and_peer) self-loop transition,
        which might occur if we have too many subscriptions for that to be possible, we force the suback response to
        be split into multiple messages.

        In practice these might be corner cases that rarely or never occur, since by default all subacks will come and
        one message and we should not receive any messages before subscribing.
        """
        async with self.CTH(add_child=True) as h:
            child = h.child
            child_subscriptions = child.mqtt_subscriptions(child.upstream_client)
            if len(child_subscriptions) < 2:
                if h.warn_if_multi_subscription_tests_skipped:
                    warnings.warn(
                        f"Skipping <{request.node.name}> because configured child proactor <{child.name}> "
                        f"has < 2 subscriptions. Subscriptions: {child_subscriptions}"
                    )
                return
            stats = child.stats.link(child.upstream_client)
            comm_event_counts = stats.comm_event_counts
            link = child._links.link(child.upstream_client)

            # unstarted child
            assert stats.num_received == 0
            assert link.state == StateName.not_started

            # start child
            child.split_client_subacks(child.upstream_client)
            child.pause_subacks()
            h.start_child()
            await await_for(
                lambda: len(child.pending_subacks) == 3,
                3,
                "ERROR waiting link reconnect",
                err_str_f=child.summary_str,
            )
            assert link.state == StateName.awaiting_setup_and_peer
            assert not link.active_for_recv()
            assert not link.active()
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 0
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert len(stats.comm_events) == 1
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child._event_persister

            # Allow one suback at a time to arrive
            # suback 1/3
            # (mqtt_suback -> awaiting_setup_and_peer)
            num_subacks = child.stats.num_received_by_type["mqtt_suback"]
            child.release_subacks(1)
            exp_subacks = num_subacks + 1
            await await_for(
                lambda: child.stats.num_received_by_type["mqtt_suback"] == exp_subacks,
                1,
                f"ERROR waiting mqtt_suback {exp_subacks} (1/3)",
                err_str_f=child.summary_str,
            )
            assert link.state == StateName.awaiting_setup_and_peer

            # suback 2/3
            # (mqtt_suback -> awaiting_setup_and_peer)
            child.release_subacks(1)
            exp_subacks += 1
            await await_for(
                lambda: child.stats.num_received_by_type["mqtt_suback"] == exp_subacks,
                1,
                f"ERROR waiting mqtt_suback {exp_subacks} (2/3)",
                err_str_f=child.summary_str,
            )
            assert link.state == StateName.awaiting_setup_and_peer

            # suback 3/3
            # (mqtt_suback -> awaiting_peer)
            child.release_subacks(1)
            exp_subacks += 1
            await await_for(
                lambda: child.stats.num_received_by_type["mqtt_suback"] == exp_subacks,
                1,
                f"ERROR waiting mqtt_suback {exp_subacks} (3/3)",
                err_str_f=child.summary_str,
            )
            assert link.state == StateName.awaiting_peer
            assert not link.active_for_recv()
            assert not link.active()
            assert link.active_for_send()
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert len(stats.comm_events) == 2
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child._event_persister

            # (message_from_peer -> awaiting_setup)
            # Tell client we lost comm
            child.pause_subacks()
            child.mqtt_client_wrapper(child.upstream_client)._client._loop_rc_handle(
                MQTT_ERR_CONN_LOST
            )
            await await_for(
                lambda: len(child.pending_subacks) == 3,
                3,
                "ERROR waiting suback pending",
                err_str_f=child.summary_str,
            )
            assert not link.active_for_send()
            assert not link.active_for_recv()
            assert not link.active()
            assert link.state == StateName.awaiting_setup_and_peer
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 2
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 1
            assert len(stats.comm_events) == 4
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child._event_persister

            # Allow one suback at a time to arrive
            # (Not strictly necessary, since message receiving code does not check if the source topic suback
            #  has arrived).
            num_subacks = child.stats.num_received_by_type["mqtt_suback"]
            child.release_subacks(1)
            exp_subacks = num_subacks + 1
            await await_for(
                lambda: child.stats.num_received_by_type["mqtt_suback"] == exp_subacks,
                1,
                f"ERROR waiting mqtt_suback {exp_subacks} (1/3)",
                err_str_f=child.summary_str,
            )
            assert link.state == StateName.awaiting_setup_and_peer

            # Start the parent, wait for it to send us a message, which will
            # transition us into awaiting_setup
            h.add_parent()
            h.start_parent()
            await await_for(
                lambda: link.in_state(StateName.awaiting_setup),
                3,
                "ERROR waiting suback pending",
                err_str_f=child.summary_str,
            )
            assert not link.active_for_send()
            assert not link.active_for_recv()
            assert not link.active()
            assert link.state == StateName.awaiting_setup
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 2
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 1
            assert len(stats.comm_events) == 4

    # @pytest.mark.skip

    @pytest.mark.asyncio
    async def test_awaiting_setup2__(self, request):
        """
        Test awaiting_setup (corner state):
         (awaiting_setup_and_peer -> message_from_peer -> awaiting_setup)
         (awaiting_setup -> mqtt_suback -> awaiting_setup)
         (awaiting_setup -> mqtt_suback -> active)
         (awaiting_setup -> message_from_peer -> awaiting_setup)
         (awaiting_setup -> disconnected -> connecting)
        Force 1 suback per subscription. By default MQTTClientWrapper packs as many subscriptions as possible into a
        single subscribe message, so by default child only receives a single suback for all subscriptions.
        So that we can test (awaiting_setup_and_peer -> mqtt_suback -> awaiting_setup_and_peer) self-loop transition,
        which might occur if we have too many subscriptions for that to be possible, we force the suback response to
        be split into multiple messages.

        In practice these might be corner cases that rarely or never occur, since by default all subacks will come and
        one message and we should not receive any messages before subscribing.
        """
        async with self.CTH(add_child=True, add_parent=True) as h:
            child = h.child
            child_subscriptions = child.mqtt_subscriptions(child.upstream_client)
            if len(child_subscriptions) < 2:
                if h.warn_if_multi_subscription_tests_skipped:
                    warnings.warn(
                        f"Skipping <{request.node.name}> because configured child proactor <{child.name}> "
                        f"has < 2 subscriptions. Subscriptions: {child_subscriptions}"
                    )
                return
            stats = child.stats.link(child.upstream_client)
            comm_event_counts = stats.comm_event_counts
            link = child._links.link(child.upstream_client)

            parent = h.parent

            def _err_str() -> str:
                return (
                    f"\nCHILD\n{child.summary_str()}\n"
                    f"\nPARENT\n{parent.summary_str()}\n"
                )

            # unstarted child
            assert stats.num_received == 0
            assert link.state == StateName.not_started

            # start child
            # (not_started -> started -> connecting)
            # (connecting -> connected -> awaiting_setup_and_peer)
            child.split_client_subacks(child.upstream_client)
            child.pause_subacks()
            h.start_child()
            await await_for(
                lambda: len(child.pending_subacks) == 3,
                3,
                "ERROR waiting link reconnect",
                err_str_f=_err_str,
            )
            assert link.state == StateName.awaiting_setup_and_peer
            assert not link.active_for_recv()
            assert not link.active()
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 0
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert len(stats.comm_events) == 1
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child._event_persister

            # Allow one suback at a time to arrive
            # (Not strictly necessary, since message receiving code does not check if the source topic suback
            #  has arrived).
            num_subacks = child.stats.num_received_by_type["mqtt_suback"]
            child.release_subacks(1)
            exp_subacks = num_subacks + 1
            await await_for(
                lambda: child.stats.num_received_by_type["mqtt_suback"] == exp_subacks,
                1,
                f"ERROR waiting mqtt_suback {exp_subacks} (1/3)",
                err_str_f=_err_str,
            )
            assert link.state == StateName.awaiting_setup_and_peer

            # (awaiting_setup_and_peer -> message_from_peer -> awaiting_setup)
            # Start the parent, wait for it to send us a message, which will
            # transition us into awaiting_setup
            h.start_parent()

            await await_for(
                lambda: link.in_state(StateName.awaiting_setup),
                3,
                "ERROR waiting suback pending",
                err_str_f=_err_str,
            )
            assert not link.active_for_send()
            assert not link.active_for_recv()
            assert not link.active()
            assert link.state == StateName.awaiting_setup
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 0
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert len(stats.comm_events) == 1

            # (awaiting_setup -> mqtt_suback -> awaiting_setup)
            # Allow another suback to arrive, remaining in awaiting_setup
            child.release_subacks(1)
            exp_subacks = num_subacks + 1
            await await_for(
                lambda: child.stats.num_received_by_type["mqtt_suback"] == exp_subacks,
                1,
                f"ERROR waiting mqtt_suback {exp_subacks} (2/3)",
                err_str_f=_err_str,
            )
            assert link.state == StateName.awaiting_setup

            # (awaiting_setup -> message_from_peer -> awaiting_setup)
            # Receive another message from peer, remaining in awaiting_setup
            dbg_topic = MQTTTopic.encode(
                "gw", parent.publication_name, DBGPayload.__fields__["TypeName"].default
            )
            assert stats.num_received_by_topic[dbg_topic] == 0
            parent.send_dbg_to_peer()
            await await_for(
                lambda: stats.num_received_by_topic[dbg_topic] == 1,
                1,
                "ERROR waiting for dbg message",
                err_str_f=_err_str,
            )
            assert link.state == StateName.awaiting_setup

            # (awaiting_setup -> disconnected -> connecting)
            # Tell client we lost comm
            child.pending_subacks.clear()
            child.pause_subacks()
            child.mqtt_client_wrapper(child.upstream_client)._client._loop_rc_handle(
                MQTT_ERR_CONN_LOST
            )
            await await_for(
                lambda: len(child.pending_subacks) == 3,
                3,
                "ERROR waiting suback pending",
                err_str_f=_err_str,
            )
            assert link.state == StateName.awaiting_setup_and_peer
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 2
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 0
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 1
            assert len(stats.comm_events) == 3
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child._event_persister

            # Allow one suback at a time to arrive
            # (Not strictly necessary, since message receiving code does not check if the source topic suback
            #  has arrived).
            num_subacks = child.stats.num_received_by_type["mqtt_suback"]
            child.release_subacks(1)
            exp_subacks = num_subacks + 1
            await await_for(
                lambda: child.stats.num_received_by_type["mqtt_suback"] == exp_subacks,
                1,
                f"ERROR waiting mqtt_suback {exp_subacks} (1/3)",
                err_str_f=_err_str,
            )
            assert link.state == StateName.awaiting_setup_and_peer

            # (awaiting_setup_and_peer -> message_from_peer -> awaiting_setup)
            # Force parent to restore comm, delivering a message, sending us to awaiting_setup
            parent.mqtt_client_wrapper(
                parent.primary_peer_client
            )._client._loop_rc_handle(MQTT_ERR_CONN_LOST)
            await await_for(
                lambda: link.in_state(StateName.awaiting_setup),
                3,
                f"ERROR waiting for message from peer",
                err_str_f=_err_str,
            )

            # (awaiting_setup -> mqtt_suback -> active)
            # Release all subacks, allowing child to go active
            child.release_subacks()
            await await_for(
                lambda: link.in_state(StateName.active),
                1,
                f"ERROR waiting for active",
                err_str_f=_err_str,
            )

    @pytest.mark.asyncio
    async def test_response_timeout(self):
        """
        Test:
            (awaiting_peer -> response_timeout -> awaiting_peer)
            (active -> response_timeout -> awaiting_peer)
        """

        async with self.CTH(
            add_child=True,
            add_parent=True,
        ) as h:
            child = h.child
            link = child._links.link(child.upstream_client)
            stats = child.stats.link(child.upstream_client)
            parent = h.parent
            parent_link = parent._links.link(parent.primary_peer_client)

            # Timeout while awaiting setup
            # (awaiting_peer -> response_timeout -> awaiting_peer)

            # start parent
            parent.pause_acks()
            h.start_parent()
            await await_for(
                lambda: parent_link.in_state(StateName.awaiting_peer),
                3,
                "ERROR waiting for parent to connect to broker",
                err_str_f=parent.summary_str,
            )

            # start child
            child.set_ack_timeout_seconds(1)
            assert stats.timeouts == 0
            h.start_child()
            await await_for(
                lambda: link.in_state(StateName.awaiting_peer),
                3,
                "ERROR waiting for child to connect to broker",
                err_str_f=parent.summary_str,
            )
            # (awaiting_peer -> response_timeout -> awaiting_peer)
            await await_for(
                lambda: stats.timeouts > 0,
                1,
                "ERROR waiting for child to timeout",
                err_str_f=parent.summary_str,
            )
            assert link.state == StateName.awaiting_peer
            assert child._event_persister.num_pending > 0

            # release the hounds
            # (awaiting_peer -> message_from_peer -> active)
            parent.release_acks()
            await await_for(
                lambda: link.in_state(StateName.active),
                1,
                "ERROR waiting for parent to restore link #1",
                err_str_f=parent.summary_str,
            )
            # wait for all events to be acked
            await await_for(
                lambda: child._event_persister.num_pending == 0,
                1,
                "ERROR waiting for events to be acked",
                err_str_f=child.summary_str,
            )

            # Timeout while active
            # (active -> response_timeout -> awaiting_peer)
            parent.pause_acks()
            child.ping_peer()
            exp_timeouts = stats.timeouts + len(
                child._links._acks._acks[child.upstream_client]
            )
            await await_for(
                lambda: stats.timeouts == exp_timeouts,
                1,
                "ERROR waiting for child to timeout",
                err_str_f=child.summary_str,
            )
            assert link.state == StateName.awaiting_peer
            assert child._event_persister.num_pending > 0
            await await_for(
                lambda: len(parent.needs_ack) == 2,
                1,
                "ERROR waiting for child to timeout",
                err_str_f=child.summary_str,
            )

            # (awaiting_peer -> message_from_peer -> active)
            parent.release_acks()
            await await_for(
                lambda: link.in_state(StateName.active),
                1,
                "ERROR waiting for parent to restore link #1",
                err_str_f=parent.summary_str,
            )

    @pytest.mark.asyncio
    async def test_ping(self):
        """
        Test:
            ping sent peridoically if no messages sent
            ping not sent if messages are sent
            ping restores comm
        """
        child_settings = self.CTH.child_settings_t()
        parent_settings = self.CTH.parent_settings_t()
        child_settings.mqtt_link_poll_seconds = (
            parent_settings.mqtt_link_poll_seconds
        ) = 0.1
        async with self.CTH(
            add_child=True,
            add_parent=True,
            child_settings=child_settings,
            parent_settings=parent_settings,
            verbose=False,
        ) as h:
            parent = h.parent
            parent_stats = parent.stats.link(parent.primary_peer_client)
            parent_ping_topic = MQTTTopic.encode(
                "gw", parent.publication_name, "gridworks-ping"
            )

            child = h.child
            child.disable_derived_events()
            child.set_ack_timeout_seconds(1)
            link = child._links.link(child.upstream_client)
            stats = child.stats.link(child.upstream_client)
            child_ping_topic = MQTTTopic.encode(
                "gw", child.publication_name, "gridworks-ping"
            )

            # start parent and child
            h.start_parent()
            h.start_child()
            await await_for(
                lambda: link.in_state(StateName.active),
                3,
                "ERROR waiting for child active",
                err_str_f=child.summary_str,
            )

            # Test that ping sent peridoically if no messages sent
            start_pings_from_parent = stats.num_received_by_topic[parent_ping_topic]
            start_pings_from_child = parent_stats.num_received_by_topic[
                child_ping_topic
            ]
            start_messages_from_parent = stats.num_received
            start_messages_from_child = parent_stats.num_received
            wait_seconds = 0.5
            await asyncio.sleep(wait_seconds)
            pings_from_parent = (
                stats.num_received_by_topic[parent_ping_topic] - start_pings_from_parent
            )
            pings_from_child = (
                parent_stats.num_received_by_topic[child_ping_topic]
                - start_pings_from_child
            )
            messages_from_parent = stats.num_received - start_messages_from_parent
            messages_from_child = parent_stats.num_received - start_messages_from_child
            exp_pings_nominal = (
                wait_seconds / parent.settings.mqtt_link_poll_seconds
            ) - 1
            err_str = (
                f"pings_from_parent: {pings_from_parent}\n"
                f"messages_from_parent: {messages_from_parent}\n"
                f"pings_from_child: {pings_from_child}\n"
                f"messages_from_child: {messages_from_child}\n"
                f"exp_pings_nominal: {exp_pings_nominal}\n"
                f"\n{child.summary_str()}\n"
                f"\n{parent.summary_str()}\n"
            )
            assert (pings_from_child + pings_from_parent) >= exp_pings_nominal, err_str
            assert messages_from_child >= exp_pings_nominal, err_str
            assert messages_from_parent >= exp_pings_nominal, err_str

            # Test that ping not sent peridoically if messages are sent
            start_pings_from_parent = stats.num_received_by_topic[parent_ping_topic]
            start_pings_from_child = parent_stats.num_received_by_topic[
                child_ping_topic
            ]
            start_messages_from_parent = stats.num_received
            start_messages_from_child = parent_stats.num_received
            reps = 50
            for _ in range(reps):
                parent.send_dbg_to_peer()
                await asyncio.sleep(0.01)
            pings_from_parent = (
                stats.num_received_by_topic[parent_ping_topic] - start_pings_from_parent
            )
            pings_from_child = (
                parent_stats.num_received_by_topic[child_ping_topic]
                - start_pings_from_child
            )
            messages_from_parent = stats.num_received - start_messages_from_parent
            messages_from_child = parent_stats.num_received - start_messages_from_child
            exp_pings_nominal = 2
            err_str = (
                f"pings_from_parent: {pings_from_parent}\n"
                f"messages_from_parent: {messages_from_parent}\n"
                f"pings_from_child: {pings_from_child}\n"
                f"messages_from_child: {messages_from_child}\n"
                f"exp_pings_nominal: {exp_pings_nominal}\n"
                f"\n{child.summary_str()}\n"
                f"\n{parent.summary_str()}\n"
            )
            assert pings_from_parent <= exp_pings_nominal, err_str
            assert pings_from_child <= exp_pings_nominal, err_str
            # Allow wide variance in number of messages exchanged - we are really testing pings, which
            # Should be should be close to 0 when a lot of messages are being exchanged.
            assert messages_from_parent >= reps * 0.5, err_str
            assert messages_from_child >= reps * 0.5, err_str

            parent.pause_acks()
            await await_for(
                lambda: link.in_state(StateName.awaiting_peer),
                child._links._acks._default_delay_seconds + 1,
                "ERROR waiting for for parent to be slow",
                err_str_f=child.summary_str,
            )
            parent.release_acks(clear=True)
            await await_for(
                lambda: link.in_state(StateName.active),
                1,
                "ERROR waiting for parent to respond",
                err_str_f=child.summary_str,
            )

    @pytest.mark.asyncio
    async def test_reupload_basic(self):
        """
        Test:
            reupload not requiring flow control
        """
        async with self.CTH(
            start_child=True,
            add_parent=True,
            verbose=False,
        ) as h:
            child = h.child
            child.disable_derived_events()
            upstream_link = h.child._links.link(child.upstream_client)
            reupload_counts = h.child.stats.link(child.upstream_client).reupload_counts
            await await_for(
                lambda: child.mqtt_quiescent(),
                1,
                "ERROR waiting for child to connect to mqtt",
                err_str_f=h.summary_str,
            )
            # Some events should have been generated, and they should have all been sent
            assert child._links.num_pending > 0
            assert child._links.num_reupload_pending == 0
            assert child._links.num_reuploaded_unacked == 0
            assert not child._links.reuploading()
            assert reupload_counts.started == 0
            assert reupload_counts.completed == 0

            # Start parent, wait for reconnect.
            h.start_parent()
            await await_for(
                lambda: upstream_link.active(),
                1,
                "ERROR waiting for parent",
                err_str_f=h.summary_str,
            )

            # Wait for reuploading to complete
            await await_for(
                lambda: reupload_counts.completed > 0,
                1,
                "ERROR waiting for re-upload to complete",
                err_str_f=h.summary_str,
            )

            # All events should have been reuploaded.
            assert child._links.num_reupload_pending == 0
            assert child._links.num_reuploaded_unacked == 0
            assert not child._links.reuploading()

    @pytest.mark.asyncio
    async def test_reupload_flow_control_simple(self):
        """
        Test:
            reupload requiring flow control
        """
        async with self.CTH(
            start_child=True,
            add_parent=True,
            child_settings=self.CTH.child_settings_t(num_initial_event_reuploads=5),
            verbose=False,
        ) as h:
            child = h.child
            child.disable_derived_events()
            upstream_link = h.child._links.link(child.upstream_client)
            reupload_counts = h.child.stats.link(child.upstream_client).reupload_counts
            await await_for(
                lambda: child.mqtt_quiescent(),
                1,
                "ERROR waiting for child to connect to mqtt",
                err_str_f=h.summary_str,
            )
            # Some events should have been generated, and they should have all been sent
            base_num_pending = child._links.num_pending
            assert base_num_pending > 0
            assert child._links.num_reupload_pending == 0
            assert child._links.num_reuploaded_unacked == 0
            assert not child._links.reuploading()

            # Generate more events than fit in pipe.
            events_to_generate = child.settings.num_initial_event_reuploads * 2
            for i in range(events_to_generate):
                child.generate_event(
                    DBGEvent(
                        Command=DBGPayload(), Msg=f"event {i+1} / {events_to_generate}"
                    )
                )
            child.logger.info(
                f"Generated {events_to_generate} events. Total pending events: {child._links.num_pending}"
            )

            # Start parent, wait for connect.
            h.start_parent()
            await await_for(
                lambda: upstream_link.active(),
                1,
                "ERROR waiting for parent",
                err_str_f=h.summary_str,
            )

            # Wait for reupload to complete
            await await_for(
                lambda: reupload_counts.completed > 0,
                1,
                "ERROR waiting for reupload to complete",
                err_str_f=h.summary_str,
            )

    @pytest.mark.asyncio
    async def test_reupload_flow_control_detail(self):
        """
        Test:
            reupload requiring flow control
        """
        async with self.CTH(
            start_child=True,
            add_parent=True,
            child_settings=self.CTH.child_settings_t(num_initial_event_reuploads=5),
            verbose=False,
            # parent_on_screen=True,
        ) as h:
            child = h.child
            child.disable_derived_events()
            child_links = h.child._links
            upstream_link = child_links.link(child.upstream_client)
            await await_for(
                lambda: child.mqtt_quiescent(),
                1,
                "ERROR waiting for child to connect to mqtt",
                err_str_f=h.summary_str,
            )
            # Some events should happened already, through the startup and mqtt connect process, and they should have
            # all been sent.
            # These events include: There are at least 3 non-generated events: startup, (mqtt connect, mqtt subscribed)/mqtt client.
            base_num_pending = child_links.num_pending
            assert base_num_pending > 0
            assert child_links.num_reupload_pending == 0
            assert child_links.num_reuploaded_unacked == 0
            assert not child_links.reuploading()

            # Generate more events than fit in pipe.
            events_to_generate = child.settings.num_initial_event_reuploads * 2
            for i in range(events_to_generate):
                child.generate_event(
                    DBGEvent(
                        Command=DBGPayload(), Msg=f"event {i+1} / {events_to_generate}"
                    )
                )
            child.logger.info(
                f"Generated {events_to_generate} events. Total pending events: {child_links.num_pending}"
            )
            assert child_links.num_reupload_pending == 0
            assert child_links.num_reuploaded_unacked == 0
            assert not child_links.reuploading()

            # pause parent acks so that we watch flow control
            h.parent.pause_acks()

            # Start parent, wait for parent to be subscribed.
            h.start_parent()
            await await_for(
                lambda: h.parent._links.link_state(h.parent.primary_peer_client)
                == StateName.awaiting_peer,
                1,
                "ERROR waiting for parent awaiting_peer",
                err_str_f=h.summary_str,
            )

            # Wait for parent to have ping waiting to be sent
            await await_for(
                lambda: len(h.parent._links.needs_ack) > 0,  # noqa
                1,
                "ERROR waiting for parent awaiting_peer",
                err_str_f=h.summary_str,
            )

            # release the ping
            h.parent.release_acks(num_to_release=1)

            # wait for child to receive ping
            await await_for(
                lambda: upstream_link.active(),
                1,
                "ERROR waiting for child peer_active",
                err_str_f=h.summary_str,
            )
            # There are 3 non-generated events: startup, mqtt connect, mqtt subscribed.
            # A "PeerActive" event is also pending but that is _not_ part of re-upload because it is
            # generated _after_ the peer is active (and therefore has its own ack timeout running, so does not need to
            # be managed by reupload).
            last_num_to_reupload = events_to_generate + base_num_pending
            last_num_reuploaded_unacked = child.settings.num_initial_event_reuploads
            last_num_repuload_pending = (
                last_num_to_reupload - child_links.num_reuploaded_unacked
            )
            err_s = (
                f"child_links.num_reuploaded_unacked: {child_links.num_reuploaded_unacked}\n"
                f"last_num_reuploaded_unacked:        {last_num_reuploaded_unacked}\n"
                f"child_links.num_reupload_pending:   {child_links.num_reupload_pending}\n"
                f"last_num_repuload_pending:          {last_num_repuload_pending}\n"
                f"{child.summary_str()}"
            )
            assert (
                child_links.num_reuploaded_unacked == last_num_reuploaded_unacked
            ), err_s
            assert child_links.num_reupload_pending == last_num_repuload_pending, err_s
            assert child_links.num_pending == last_num_to_reupload + 1
            assert child_links.reuploading()

            parent_ack_topic = MQTTTopic.encode(
                "gw", h.parent.publication_name, "gridworks-ack"
            )
            acks_received_by_child = child.stats.num_received_by_topic[parent_ack_topic]

            # Release acks one by one.
            #
            #   Bound this loop by time, not by total number of acks since at least one non-reupload ack should arrive
            #   (for the PeerActive event) and others could arrive if, for example, a duplicate MQTT message appeared.
            #
            end_time = time.time() + 5
            # loop_count_dbg = 0
            acks_released = 0
            while child_links.reuploading() and time.time() < end_time:
                # loop_path_dbg = 0
                # loop_count_dbg += 1

                # release one ack
                acks_released += h.parent.release_acks(num_to_release=1)

                # Wait for child to receive an ack
                await await_for(
                    lambda: child.stats.num_received_by_topic[parent_ack_topic]
                    == acks_received_by_child + acks_released,
                    1,
                    f"ERROR waiting for child to receive ack (acks_released: {acks_released})",
                    err_str_f=h.summary_str,
                )
                curr_num_reuploaded_unacked = child_links.num_reuploaded_unacked
                curr_num_repuload_pending = child_links.num_reupload_pending
                curr_num_to_reuplad = (
                    curr_num_reuploaded_unacked + curr_num_repuload_pending
                )
                if curr_num_to_reuplad == last_num_to_reupload:
                    # loop_path_dbg |= 0x00000001
                    assert curr_num_reuploaded_unacked == last_num_reuploaded_unacked
                    assert curr_num_repuload_pending == last_num_repuload_pending
                elif curr_num_to_reuplad == last_num_to_reupload - 1:
                    # loop_path_dbg |= 0x00000002
                    if curr_num_reuploaded_unacked == last_num_reuploaded_unacked:
                        assert (
                            curr_num_repuload_pending == last_num_repuload_pending - 1
                        )
                    else:
                        assert (
                            curr_num_reuploaded_unacked
                            == last_num_reuploaded_unacked - 1
                        )
                        assert curr_num_repuload_pending == last_num_repuload_pending
                    assert child_links.reuploading() == bool(
                        curr_num_reuploaded_unacked > 0
                    )
                else:
                    raise ValueError(
                        "Unexpected change in reupload counts: "
                        f"({last_num_reuploaded_unacked}, {last_num_repuload_pending}) -> "
                        f"({curr_num_reuploaded_unacked}, {curr_num_repuload_pending})"
                    )

                # child.logger.info(
                #     f"ack loop: {loop_count_dbg} / {acks_released}:"
                #     f"({last_num_reuploaded_unacked}, {last_num_repuload_pending}) -> "
                #     f"({curr_num_reuploaded_unacked}, {curr_num_repuload_pending})"
                #     f" loop_path_dbg: 0x{loop_path_dbg:08X}")

                last_num_to_reupload = curr_num_to_reuplad
                last_num_reuploaded_unacked = curr_num_reuploaded_unacked
                last_num_repuload_pending = curr_num_repuload_pending

            assert not child_links.reuploading()

    @pytest.mark.asyncio
    async def test_reupload_errors(self):
        async with self.CTH(
            start_child=True,
            add_parent=True,
            child_verbose=False,
        ) as h:
            child = h.child
            child.disable_derived_events()
            reupload_counts = h.child.stats.link(child.upstream_client).reupload_counts
            child_links = h.child._links
            upstream_link = child_links.link(child.upstream_client)
            parent = h.parent

            def _err_str() -> str:
                return (
                    f"\nCHILD\n{child.summary_str()}\n"
                    f"\nPARENT\n{parent.summary_str()}\n"
                )

            await await_for(
                lambda: child.mqtt_quiescent(),
                1,
                "ERROR waiting for child to connect to mqtt",
                err_str_f=_err_str,
            )
            base_num_pending = child_links.num_pending
            assert base_num_pending > 0
            assert child_links.num_reupload_pending == 0
            assert child_links.num_reuploaded_unacked == 0
            assert not child_links.reuploading()

            generator = _EventGen(child)
            generator.generate(num_corrupt=10)
            generator.generate(num_ok=10)
            generator.generate(num_empty=10)
            generator.generate(num_ok=10)
            generator.generate(num_missing=10)
            generator.generate(num_ok=10)

            h.start_parent()
            await await_for(
                lambda: upstream_link.active(),
                1,
                "ERROR waiting for active",
                err_str_f=_err_str,
            )

            # Wait for reupload to complete
            await await_for(
                lambda: reupload_counts.completed > 0,
                3,
                "ERROR waiting for reupload to complete",
                err_str_f=_err_str,
            )
            assert reupload_counts.started == reupload_counts.completed
            assert parent.stats.num_events_received >= base_num_pending + 60



from gwproactor_test.certs import TEST_CA_CERTIFICATE_PATH_VAR
from gwproactor_test.certs import TEST_CA_PRIVATE_KEY_VAR
from gwproactor_test.certs import TEST_CERTIFICATE_CACHE_VAR
from gwproactor_test.certs import copy_keys
from gwproactor_test.certs import set_test_certificate_cache_dir
from gwproactor_test.certs import test_ca_certificate_path
from gwproactor_test.certs import test_ca_private_key_path
from gwproactor_test.certs import test_certificate_cache_dir
from gwproactor_test.clean import DefaultTestEnv
from gwproactor_test.clean import clean_test_env
from gwproactor_test.clean import default_test_env
from gwproactor_test.comm_test_helper import CommTestHelper
from gwproactor_test.comm_test_helper import ProactorTestHelper
from gwproactor_test.logger_guard import LoggerGuard
from gwproactor_test.logger_guard import LoggerGuards
from gwproactor_test.logger_guard import restore_loggers
from gwproactor_test.proactor_recorder import ProactorT
from gwproactor_test.proactor_recorder import RecorderInterface
from gwproactor_test.proactor_recorder import RecorderLinkStats
from gwproactor_test.proactor_recorder import RecorderStats
from gwproactor_test.proactor_recorder import make_recorder_class
from gwproactor_test.proactor_test_collections import ProactorCommTests
from gwproactor_test.wait import AwaitablePredicate
from gwproactor_test.wait import ErrorStringFunction
from gwproactor_test.wait import Predicate
from gwproactor_test.wait import StopWatch
from gwproactor_test.wait import await_for


__all__ = [
    "TEST_CERTIFICATE_CACHE_VAR",
    "TEST_CA_PRIVATE_KEY_VAR",
    "TEST_CA_CERTIFICATE_PATH_VAR",
    "test_ca_certificate_path",
    "test_ca_private_key_path",
    "copy_keys",
    "set_test_certificate_cache_dir",
    "test_certificate_cache_dir",
    "DefaultTestEnv",
    "clean_test_env",
    "default_test_env",
    "CommTestHelper",
    "ProactorTestHelper",
    "LoggerGuard",
    "LoggerGuards",
    "restore_loggers",
    "ProactorT",
    "RecorderInterface",
    "RecorderLinkStats",
    "RecorderStats",
    "make_recorder_class",
    "ProactorCommTests",
    "AwaitablePredicate",
    "ErrorStringFunction",
    "Predicate",
    "StopWatch",
    "await_for",
]



"""Proactor interfaces, separate from implementations to clarify how users of this package interact with it and to
create forward references for implementation hiearchies
"""

import asyncio
import importlib
import sys
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from typing import Any
from typing import Coroutine
from typing import Optional
from typing import Sequence
from typing import Type
from typing import TypeVar

from aiohttp.typedefs import Handler as HTTPHandler
from gwproto import HardwareLayout
from gwproto import ShNode
from gwproto.messages import EventT
from result import Result

from gwproactor.config.proactor_settings import ProactorSettings
from gwproactor.external_watchdog import ExternalWatchdogCommandBuilder
from gwproactor.logger import ProactorLogger
from gwproactor.message import Message
from gwproactor.stats import ProactorStats


T = TypeVar("T")


@dataclass
class MonitoredName:
    name: str
    timeout_seconds: float


class CommunicatorInterface(ABC):
    """Pure interface necessary for interaction between a sub-object and the system services proactor"""

    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def _send(self, message: Message):
        raise NotImplementedError

    @abstractmethod
    def process_message(self, message: Message) -> Result[bool, BaseException]:
        raise NotImplementedError

    @property
    @abstractmethod
    def monitored_names(self) -> Sequence[MonitoredName]:
        raise NotImplementedError

    @property
    @abstractmethod
    def services(self) -> "ServicesInterface":
        raise NotImplementedError


class Communicator(CommunicatorInterface, ABC):
    """A partial implementation of CommunicatorInterface which supplies the trivial implementations"""

    _name: str
    _services: "ServicesInterface"

    def __init__(self, name: str, services: "ServicesInterface"):
        self._name = name
        self._services = services

    @property
    def name(self) -> str:
        return self._name

    def _send(self, message: Message) -> None:
        self._services.send(message)

    @property
    def monitored_names(self) -> Sequence[MonitoredName]:
        return []

    @property
    def services(self) -> "ServicesInterface":
        return self._services


class Runnable(ABC):
    """Pure interface to an object which is expected to support starting, stopping and joining."""

    @abstractmethod
    def start(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def stop(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def join(self) -> None:
        raise NotImplementedError

    async def stop_and_join(self) -> None:
        self.stop()
        await self.join()


class ActorInterface(CommunicatorInterface, Runnable, ABC):
    """Pure interface for a proactor sub-object (an Actor) which can communicate
    and has a GridWorks ShNode."""

    @property
    @abstractmethod
    def alias(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def node(self) -> ShNode:
        raise NotImplementedError

    @abstractmethod
    def init(self) -> None:
        """Called after constructor so derived functions can be used in setup."""

    @classmethod
    def load(
        cls,
        name: str,
        actor_class_name: str,
        services: "ServicesInterface",
        module_name: str,
    ) -> "ActorInterface":
        if module_name not in sys.modules:
            importlib.import_module(module_name)
        actor_class = getattr(sys.modules[module_name], actor_class_name)
        actor = actor_class(name, services)
        actor.init()
        return actor


INVALID_IO_TASK_HANDLE = -1


class IOLoopInterface(CommunicatorInterface, Runnable, ABC):
    """Interface to an asyncio event loop running a seperate thread meant io-only
    routines which have minimal CPU bound work."""

    @abstractmethod
    def add_io_coroutine(self, coro: Coroutine, name: str = "") -> int:
        """
        Add a couroutine that will be run as a task in the io event loop.

        May be called before or after IOLoopInterface.start(). No tasks will actually
        run until IOLoopInterface.start() is called.

        This routine is thread safe.

        Args:
            coro: The coroutine to run as a task.
            name: Optional name assigned to task for use in debugging.

        Returns:
            an integer handle which may be passed to cancel_io_coroutine() to cancel
            the task running the coroutine.
        """

    @abstractmethod
    def cancel_io_routine(self, handle: int) -> None:
        """Cancel the task represented by the handle.

        This routine may be called multiple times for the same handle with no effect.
        This routine is thread safe.

        Args:
            handle: The handle returned by previous call to add_io_routine().
        """


class ServicesInterface(CommunicatorInterface):
    """Interface to system services (the proactor)"""

    @abstractmethod
    def get_communicator(self, name: str) -> Optional[CommunicatorInterface]:
        raise NotImplementedError

    @abstractmethod
    def get_communicator_as_type(self, name: str, type_: Type[T]) -> Optional[T]:
        raise NotImplementedError

    @abstractmethod
    def send(self, message: Message) -> None:
        raise NotImplementedError

    @abstractmethod
    def send_threadsafe(self, message: Message) -> None:
        raise NotImplementedError

    @property
    @abstractmethod
    def async_receive_queue(self) -> Optional[asyncio.Queue]:
        raise NotImplementedError

    @property
    @abstractmethod
    def event_loop(self) -> Optional[asyncio.AbstractEventLoop]:
        raise NotImplementedError

    @property
    @abstractmethod
    def io_loop_manager(self) -> IOLoopInterface:
        raise NotImplementedError

    @abstractmethod
    def add_web_server_config(
        self, name: str, host: str, port: int, **kwargs: Any
    ) -> None:
        """Adds configuration for web server which will be started when start() is called.

        Not thread safe."""
        raise NotImplementedError

    @abstractmethod
    def add_web_route(
        self,
        server_name: str,
        method: str,
        path: str,
        handler: HTTPHandler,
        **kwargs: Any
    ):
        """Adds configuration for web server route which will be available after start() is called.

        May be called even if associated web server is not configured, in which case this route
        will simply be ignored.

        Not thread safe.
        """
        raise NotImplementedError

    @abstractmethod
    def generate_event(self, event: EventT) -> None:
        raise NotImplementedError

    @property
    @abstractmethod
    def publication_name(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def settings(self) -> ProactorSettings:
        raise NotImplementedError

    @property
    @abstractmethod
    def logger(self) -> ProactorLogger:
        raise NotImplementedError

    @property
    @abstractmethod
    def stats(self) -> ProactorStats:
        raise NotImplementedError

    @property
    @abstractmethod
    def hardware_layout(self) -> HardwareLayout: ...

    @abstractmethod
    def get_external_watchdog_builder_class(
        self,
    ) -> type[ExternalWatchdogCommandBuilder]:
        raise NotImplementedError



from typing import Any
from typing import Optional

from pydantic import root_validator
from pydantic import validator

from gwproactor import ProactorSettings
from gwproactor.config import LoggingSettings
from gwproactor.config import MQTTClient
from gwproactor.config import Paths
from gwproactor_test.dummies.names import DUMMY_PARENT_ENV_PREFIX
from gwproactor_test.dummies.names import DUMMY_PARENT_NAME


class DummyParentSettings(ProactorSettings):
    child_mqtt: MQTTClient = MQTTClient()

    class Config(ProactorSettings.Config):
        env_prefix = DUMMY_PARENT_ENV_PREFIX

    @root_validator(pre=True)
    def pre_root_validator(cls, values: dict) -> dict:
        return ProactorSettings.update_paths_name(values, DUMMY_PARENT_NAME)



"""Proactor implementation"""

import asyncio
import sys
import traceback
import uuid
from typing import Any
from typing import Awaitable
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Type
from typing import TypeVar

import gwproto
from aiohttp.typedefs import Handler as HTTPHandler
from gwproto.data_classes.components.web_server_component import WebServerComponent
from gwproto.data_classes.hardware_layout import HardwareLayout
from gwproto.data_classes.sh_node import ShNode
from gwproto.messages import Ack
from gwproto.messages import EventBase
from gwproto.messages import EventT
from gwproto.messages import Ping
from gwproto.messages import ProblemEvent
from gwproto.messages import ShutdownEvent
from result import Err
from result import Ok
from result import Result

from gwproactor import ProactorSettings
from gwproactor.external_watchdog import ExternalWatchdogCommandBuilder
from gwproactor.external_watchdog import SystemDWatchdogCommandBuilder
from gwproactor.io_loop import IOLoop
from gwproactor.links import AckWaitInfo
from gwproactor.links import AsyncioTimerManager
from gwproactor.links import LinkManager
from gwproactor.links import LinkManagerTransition
from gwproactor.links import Transition
from gwproactor.logger import ProactorLogger
from gwproactor.message import DBGCommands
from gwproactor.message import DBGEvent
from gwproactor.message import DBGPayload
from gwproactor.message import Message
from gwproactor.message import MQTTConnectFailPayload
from gwproactor.message import MQTTConnectPayload
from gwproactor.message import MQTTDisconnectPayload
from gwproactor.message import MQTTProblemsPayload
from gwproactor.message import MQTTReceiptPayload
from gwproactor.message import MQTTSubackPayload
from gwproactor.message import PatWatchdog
from gwproactor.message import Shutdown
from gwproactor.persister import PersisterInterface
from gwproactor.persister import StubPersister
from gwproactor.proactor_interface import CommunicatorInterface
from gwproactor.proactor_interface import IOLoopInterface
from gwproactor.proactor_interface import MonitoredName
from gwproactor.proactor_interface import Runnable
from gwproactor.proactor_interface import ServicesInterface
from gwproactor.problems import Problems
from gwproactor.stats import ProactorStats
from gwproactor.str_tasks import str_tasks
from gwproactor.watchdog import WatchdogManager
from gwproactor.web_manager import _WebManager


T = TypeVar("T")


class Proactor(ServicesInterface, Runnable):

    _name: str
    _settings: ProactorSettings
    _node: ShNode
    _layout: HardwareLayout
    _logger: ProactorLogger
    _stats: ProactorStats
    _event_persister: PersisterInterface
    _reindex_problems: Optional[Problems] = None
    _loop: Optional[asyncio.AbstractEventLoop] = None
    _receive_queue: Optional[asyncio.Queue] = None
    _links: LinkManager
    _communicators: Dict[str, CommunicatorInterface]
    _stop_requested: bool
    _tasks: List[asyncio.Task]
    _io_loop_manager: IOLoop
    _web_manager: _WebManager
    _watchdog: WatchdogManager

    def __init__(
        self,
        name: str,
        settings: ProactorSettings,
        hardware_layout: Optional[HardwareLayout] = None,
    ):
        self._name = name
        self._settings = settings
        if hardware_layout is None:
            hardware_layout = HardwareLayout(
                dict(
                    ShNodes=[
                        dict(
                            ShNodeId=str(uuid.uuid4()),
                            Alias=self._name,
                            ActorClassGtEnumSymbol="00000000",
                            RoleGtEnumSymbol="00000000",
                            TypeName="spaceheat.node.gt",
                        )
                    ]
                )
            )
        self._layout = hardware_layout
        self._node = self._layout.node(name)
        self._logger = ProactorLogger(**settings.logging.qualified_logger_names())
        self._stats = self.make_stats()
        self._event_persister = self.make_event_persister(settings)
        reindex_result = self._event_persister.reindex()
        if reindex_result.is_err():
            self._reindex_problems = reindex_result.err()
            self._logger.error("ERROR in event persister reindex():")
            self._logger.error(reindex_result.err())
        self._links = LinkManager(
            publication_name=self.publication_name,
            settings=settings,
            logger=self._logger,
            stats=self._stats,
            event_persister=self._event_persister,
            timer_manager=AsyncioTimerManager(),
            ack_timeout_callback=self._process_ack_timeout,
        )
        self._communicators = dict()
        self._tasks = []
        self._stop_requested = False
        self._watchdog = WatchdogManager(9, self)
        self.add_communicator(self._watchdog)
        self._io_loop_manager = IOLoop(self)
        self.add_communicator(self._io_loop_manager)
        self._web_manager = _WebManager(self)
        self.add_communicator(self._web_manager)
        for config in self._layout.get_components_by_type(WebServerComponent):
            self._web_manager.add_web_server_config(
                name=config.web_server_gt.Name,
                host=config.web_server_gt.Host,
                port=config.web_server_gt.Port,
                **config.web_server_gt.Kwargs,
            )

    @classmethod
    def make_stats(cls) -> ProactorStats:
        return ProactorStats()

    @classmethod
    def make_event_persister(cls, settings: ProactorSettings) -> PersisterInterface:
        return StubPersister()

    def send(self, message: Message):
        if not isinstance(message.Payload, PatWatchdog):
            self._logger.message_summary(
                "OUT internal",
                message.Header.Src,
                f"{message.Header.Dst}/{message.Header.MessageType}",
                message.Payload,
                message_id=message.Header.MessageId,
            )
        self._receive_queue.put_nowait(message)

    def send_threadsafe(self, message: Message) -> None:
        self._loop.call_soon_threadsafe(self._receive_queue.put_nowait, message)

    def get_communicator(self, name: str) -> Optional[CommunicatorInterface]:
        return self._communicators.get(name, None)

    def get_communicator_as_type(self, name: str, type_: Type[T]) -> Optional[T]:
        communicator = self.get_communicator(name)
        if communicator is not None and not isinstance(communicator, type_):
            raise ValueError(
                f"ERROR. Communicator <{name}> has type {type(communicator)} not {type_}"
            )
        return communicator

    @property
    def name(self) -> str:
        return self._name

    @property
    def publication_name(self) -> str:
        return self._name

    @property
    def monitored_names(self) -> Sequence[MonitoredName]:
        return []

    @property
    def settings(self) -> ProactorSettings:
        return self._settings

    @property
    def logger(self) -> ProactorLogger:
        return self._logger

    @property
    def stats(self) -> ProactorStats:
        return self._stats

    @property
    def io_loop_manager(self) -> IOLoopInterface:
        return self._io_loop_manager

    def add_web_server_config(
        self, name: str, host: str, port: int, **kwargs: Any
    ) -> None:
        self._web_manager.add_web_server_config(
            name=name, host=host, port=port, **kwargs
        )

    def add_web_route(
        self,
        server_name: str,
        method: str,
        path: str,
        handler: HTTPHandler,
        **kwargs: Any,
    ):
        self._web_manager.add_web_route(
            server_name=server_name, method=method, path=path, handler=handler, **kwargs
        )

    @property
    def hardware_layout(self) -> HardwareLayout:
        return self._layout

    @property
    def services(self) -> "ServicesInterface":
        return self

    def get_external_watchdog_builder_class(
        self,
    ) -> type[ExternalWatchdogCommandBuilder]:
        return SystemDWatchdogCommandBuilder

    @property
    def upstream_client(self) -> str:
        return self._links.upstream_client

    @property
    def primary_peer_client(self) -> str:
        return self._links.primary_peer_client

    def _send(self, message: Message):
        self.send(message)

    def generate_event(self, event: EventT) -> Result[bool, BaseException]:
        return self._links.generate_event(event)

    def _process_ack_timeout(self, wait_info: AckWaitInfo) -> None:
        self._logger.message_enter(
            "++Proactor._process_ack_timeout %s", wait_info.message_id
        )
        path_dbg = 0
        result = self._links.process_ack_timeout(wait_info)
        if result.is_ok():
            path_dbg |= 0x00000001
            if result.value.deactivated():
                path_dbg |= 0x00000002
                self._derived_recv_deactivated(result.value)
        else:
            path_dbg |= 0x00000004
            self._report_error(result.err(), msg="Proactor._process_ack_timeout")
        self._logger.message_exit(
            "--Proactor._process_ack_timeout path:0x%08X", path_dbg
        )

    def _process_ack(self, link_name: str, message_id: str):
        self._links.process_ack(link_name, message_id)

    def _process_dbg(self, dbg: DBGPayload):
        self._logger.path("++_process_dbg")
        path_dbg = 0
        count_dbg = 0
        for logger_name in ["message_summary", "lifecycle", "comm_event"]:
            requested_level = getattr(dbg.Levels, logger_name)
            if requested_level > -1:
                path_dbg |= 0x00000001
                count_dbg += 1
                logger = getattr(self._logger, logger_name + "_logger")
                old_level = logger.getEffectiveLevel()
                logger.setLevel(requested_level)
                self._logger.debug(
                    "%s logger level %s -> %s",
                    logger_name,
                    old_level,
                    logger.getEffectiveLevel(),
                )
        match dbg.Command:
            case DBGCommands.show_subscriptions:
                path_dbg |= 0x00000002
                self._links.log_subscriptions("message")
            case _:
                path_dbg |= 0x00000004
        self.generate_event(
            DBGEvent(Command=dbg, Path=f"0x{path_dbg:08X}", Count=count_dbg, Msg="")
        )
        self._logger.path("--_process_dbg  path:0x%08X  count:%d", path_dbg, count_dbg)

    def add_communicator(self, communicator: CommunicatorInterface):
        if communicator.name in self._communicators:
            raise ValueError(
                f"ERROR. Communicator with name [{communicator.name}] already present"
            )
        self._communicators[communicator.name] = communicator
        for monitored in communicator.monitored_names:
            self._watchdog.add_monitored_name(monitored)

    @property
    def async_receive_queue(self) -> Optional[asyncio.Queue]:
        return self._receive_queue

    @property
    def event_loop(self) -> Optional[asyncio.AbstractEventLoop]:
        return self._loop

    async def process_messages(self):
        # noinspection PyBroadException
        try:
            self._start_processing_messages()
            while not self._stop_requested:
                message = await self._receive_queue.get()
                if not self._stop_requested:
                    await self.process_message(message)
                self._receive_queue.task_done()
        except BaseException as e:
            if not isinstance(e, asyncio.exceptions.CancelledError):
                self._logger.exception(f"ERROR in process_message")
                self._logger.error("Stopping proactor")
                # noinspection PyBroadException
                try:
                    self.generate_event(
                        ShutdownEvent(
                            Reason=(
                                f"ERROR in process_message {e}\n"
                                f"{traceback.format_exception(e)}"
                            )
                        )
                    )
                except:
                    self._logger.exception(f"ERROR generating exception event")
        # noinspection PyBroadException
        try:
            self.stop()
        except:
            self._logger.exception(f"ERROR stopping proactor")

    def start_tasks(self):
        self._tasks = [
            asyncio.create_task(self.process_messages(), name="process_messages"),
        ] + self._links.start_ping_tasks()
        self._start_derived_tasks()

    def _start_derived_tasks(self):
        pass

    def _derived_process_message(self, message: Message):
        pass

    def _derived_process_mqtt_message(
        self, message: Message[MQTTReceiptPayload], decoded: Any
    ):
        pass

    @classmethod
    def _second_caller(cls) -> str:
        try:
            # noinspection PyProtectedMember,PyUnresolvedReferences
            return sys._getframe(2).f_back.f_code.co_name
        except BaseException as e:
            return f"[ERROR extracting caller of _report_errors: {e}"

    def _report_error(
        self, error: BaseException, msg: str = ""
    ) -> Result[bool, BaseException]:
        try:
            if not msg:
                msg = self._second_caller()
            self._report_errors([error], msg)
        except BaseException as e2:
            return Err(e2)
        return Ok()

    def _report_errors(
        self, errors: Sequence[BaseException], msg: str = ""
    ) -> Result[bool, BaseException]:
        try:
            if not msg:
                msg = self._second_caller()
            self.generate_event(Problems(errors=errors).problem_event(msg))
        except BaseException as e2:
            return Err(e2)
        return Ok()

    def _start_processing_messages(self):
        """Hook for processing before any messages are pulled from queue"""

    async def process_message(self, message: Message):
        if not isinstance(message.Payload, PatWatchdog):
            self._logger.message_enter(
                "++Proactor.process_message %s/%s",
                message.Header.Src,
                message.Header.MessageType,
            )
        path_dbg = 0
        if not isinstance(message.Payload, (MQTTReceiptPayload, PatWatchdog)):
            path_dbg |= 0x00000001
            self._logger.message_summary(
                "IN  internal",
                self.name,
                f"{message.Header.Src}/{message.Header.MessageType}",
                message.Payload,
                message_id=message.Header.MessageId,
            )
        self._stats.add_message(message)
        match message.Payload:
            case MQTTReceiptPayload():
                path_dbg |= 0x00000002
                self._process_mqtt_message(message)
            case MQTTConnectPayload():
                path_dbg |= 0x00000004
                self._process_mqtt_connected(message)
            case MQTTDisconnectPayload():
                path_dbg |= 0x00000008
                self._process_mqtt_disconnected(message)
            case MQTTConnectFailPayload():
                path_dbg |= 0x00000010
                self._process_mqtt_connect_fail(message)
            case MQTTSubackPayload():
                path_dbg |= 0x00000020
                self._process_mqtt_suback(message)
            case MQTTProblemsPayload():
                path_dbg |= 0x00000040
                self._process_mqtt_problems(message)
            case PatWatchdog():
                path_dbg |= 0x00000080
                self._watchdog.process_message(message)
            case Shutdown():
                path_dbg |= 0x00000100
                self._process_shutdown_message(message)
            case EventBase():
                path_dbg |= 0x00000200
                self.generate_event(message.Payload)
            case _:
                path_dbg |= 0x00000400
                self._derived_process_message(message)
        if not isinstance(message.Payload, PatWatchdog):
            self._logger.message_exit(
                "--Proactor.process_message  path:0x%08X", path_dbg
            )

    def _decode_mqtt_message(self, mqtt_payload) -> Result[Message[Any], BaseException]:
        try:
            result = Ok(
                self._links.decode(
                    mqtt_payload.client_name,
                    mqtt_payload.message.topic,
                    mqtt_payload.message.payload,
                )
            )
        except Exception as e:
            self._logger.exception("ERROR decoding [%s]", mqtt_payload)
            self.generate_event(
                ProblemEvent(
                    ProblemType=gwproto.messages.Problems.warning,
                    Summary=f"Decoding error topic [{mqtt_payload.message.topic}]  error [{type(e)}]",
                    Details=(
                        f"Topic: {mqtt_payload.message.topic}\n"
                        f"Message: {mqtt_payload.message.payload[:70]}"
                        f"{'...' if len(mqtt_payload.message.payload)> 70 else ''}\n"
                        f"{traceback.format_exception(e)}\n"
                        f"Exception: {e}"
                    ),
                )
            )
            result = Err(e)
        return result

    def _process_mqtt_message(
        self, mqtt_receipt_message: Message[MQTTReceiptPayload]
    ) -> Result[Message[Any], BaseException]:
        self._logger.path(
            "++Proactor._process_mqtt_message %s/%s",
            mqtt_receipt_message.Header.Src,
            mqtt_receipt_message.Header.MessageType,
        )
        path_dbg = 0
        self._stats.add_mqtt_message(mqtt_receipt_message)
        decode_result = self._decode_mqtt_message(mqtt_receipt_message.Payload)
        if decode_result.is_ok():
            path_dbg |= 0x00000001
            decoded_message = decode_result.value
            self._logger.message_summary(
                "IN  mqtt    ",
                self.name,
                mqtt_receipt_message.Payload.message.topic,
                decoded_message.Payload,
                message_id=decoded_message.Header.MessageId,
            )
            link_mgr_results = self._links.process_mqtt_message(mqtt_receipt_message)
            if link_mgr_results.is_ok():
                path_dbg |= 0x00000002
                if link_mgr_results.value.recv_activated():
                    path_dbg |= 0x00000004
                    self._derived_recv_activated(link_mgr_results.value)
            else:
                path_dbg |= 0x00000008
                self._report_error(
                    link_mgr_results.err(),
                    "_process_mqtt_message/_link_states.process_mqtt_message",
                )
            match decoded_message.Payload:
                case Ack():
                    path_dbg |= 0x00000010
                    self._process_ack(
                        mqtt_receipt_message.Payload.client_name,
                        decoded_message.Payload.AckMessageID,
                    )
                case Ping():
                    path_dbg |= 0x00000020
                case DBGPayload():
                    path_dbg |= 0x00000040
                    self._process_dbg(decoded_message.Payload)
                case _:
                    path_dbg |= 0x00000080
                    self._derived_process_mqtt_message(
                        mqtt_receipt_message, decoded_message
                    )
            if decoded_message.Header.AckRequired:
                path_dbg |= 0x00000100
                self._links.send_ack(
                    mqtt_receipt_message.Payload.client_name, decoded_message
                )
        self._logger.path(
            "--Proactor._process_mqtt_message:%s  path:0x%08X",
            int(decode_result.is_ok()),
            path_dbg,
        )
        return decode_result

    def _process_mqtt_connected(self, message: Message[MQTTConnectPayload]):
        result = self._links.process_mqtt_connected(message)
        if result.is_err():
            self._report_error(result.err(), "_process_mqtt_connected")

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def _derived_recv_deactivated(
        self, transition: LinkManagerTransition
    ) -> Result[bool, BaseException]:
        return Ok()

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def _derived_recv_activated(
        self, transition: Transition
    ) -> Result[bool, BaseException]:
        return Ok()

    def _process_mqtt_disconnected(
        self, message: Message[MQTTDisconnectPayload]
    ) -> Result[bool, BaseException]:
        link_mgr_result = self._links.process_mqtt_disconnected(message)
        if link_mgr_result.is_ok() and link_mgr_result.value.recv_deactivated():
            result = self._derived_recv_deactivated(link_mgr_result.value)
        else:
            result = Err(link_mgr_result.err())
        return result

    def _process_mqtt_connect_fail(
        self, message: Message[MQTTConnectFailPayload]
    ) -> Result[bool, BaseException]:
        return self._links.process_mqtt_connect_fail(message)

    def _process_mqtt_suback(
        self, message: Message[MQTTSubackPayload]
    ) -> Result[bool, BaseException]:
        self._logger.path(
            "++Proactor._process_mqtt_suback client:%s", message.Payload.client_name
        )
        path_dbg = 0
        link_mgr_result = self._links.process_mqtt_suback(message)
        if link_mgr_result.is_ok():
            path_dbg |= 0x00000001
            if link_mgr_result.value.recv_activated():
                path_dbg |= 0x00000002
                result = self._derived_recv_activated(link_mgr_result.value)
            else:
                path_dbg |= 0x00000004
                result = Ok(True)
        else:
            path_dbg |= 0x00000008
            result = Err(link_mgr_result.err())
        self._logger.path(
            "--Proactor._process_mqtt_suback:%d  path:0x%08X",
            result.is_ok(),
            path_dbg,
        )
        return result

    def _process_mqtt_problems(
        self, message: Message[MQTTProblemsPayload]
    ) -> Result[bool, BaseException]:
        self.generate_event(
            ProblemEvent(
                ProblemType=gwproto.messages.Problems.error,
                Summary=f"Error in mqtt event loop for client [{message.Payload.client_name}]",
                Details=(
                    f"{message.Payload.problems}\n"
                    f"{message.Payload.problems.error_traceback_str()}"
                ),
            )
        )
        return Ok()

    def _process_shutdown_message(self, message: Message[Shutdown]):
        self._stop_requested = True
        self.generate_event(ShutdownEvent(Reason=message.Payload.Reason))
        self._logger.lifecycle(
            f"Shutting down due to ShutdownMessage, [{message.Payload.Reason}]"
        )

    async def run_forever(self):
        self._loop = asyncio.get_running_loop()
        self._receive_queue = asyncio.Queue()
        self._links.start(self._loop, self._receive_queue)
        if self._reindex_problems is not None:
            self.generate_event(
                self._reindex_problems.problem_event("Startup event reindex() problems")
            )
        self._reindex_problems = None
        for communicator in self._communicators.values():
            if isinstance(communicator, Runnable):
                communicator.start()
        self.start_tasks()
        await self.join()

    def start(self):
        raise RuntimeError("ERROR. Proactor must be started by awaiting run_forever()")

    def stop(self):
        self._stop_requested = True
        for task in self._tasks:
            # TODO: CS - Send self a shutdown message instead?
            if not task.done():
                task.cancel()
        self._links.stop()
        for communicator in self._communicators.values():
            if isinstance(communicator, Runnable):
                # noinspection PyBroadException
                try:
                    communicator.stop()
                except:
                    pass

    async def join(self):
        self._logger.lifecycle("++Proactor.join()")
        self._logger.lifecycle(str_tasks(self._loop, "Proactor.join() - all tasks"))
        running: List[Awaitable] = self._tasks[:]
        for communicator in self._communicators.values():
            communicator_name = communicator.name
            if isinstance(communicator, Runnable):
                running.append(
                    self._loop.create_task(
                        communicator.join(), name=f"{communicator_name}.join"
                    )
                )
        # noinspection PyBroadException
        try:
            while running:
                self._logger.lifecycle(
                    str_tasks(self._loop, "WAITING FOR", tasks=running)
                )
                done, running = await asyncio.wait(
                    running, return_when="FIRST_COMPLETED"
                )
                self._logger.lifecycle(str_tasks(self._loop, tag="DONE", tasks=done))
                self._logger.lifecycle(
                    str_tasks(self._loop, tag="PENDING", tasks=running)
                )
                for task in done:
                    if not task.cancelled() and (exception := task.exception()):
                        self._logger.error(
                            f"EXCEPTION in task {task.get_name()}  {exception}"
                        )
                        self._logger.error(traceback.format_tb(exception.__traceback__))
        except BaseException as e:
            self._logger.exception("ERROR in Proactor.join: %s <%s>", type(e), e)
        self._logger.lifecycle("--Proactor.join()")



import asyncio
import subprocess
import time
from typing import Optional

from gwproto import Message

from gwproactor.message import InternalShutdownMessage
from gwproactor.message import KnownNames
from gwproactor.message import PatExternalWatchdog
from gwproactor.message import PatExternalWatchdogMessage
from gwproactor.message import PatInternalWatchdog
from gwproactor.proactor_interface import Communicator
from gwproactor.proactor_interface import MonitoredName
from gwproactor.proactor_interface import Runnable
from gwproactor.proactor_interface import ServicesInterface


class _MonitoredName(MonitoredName):
    last_pat: float = 0.0


class WatchdogManager(Communicator, Runnable):
    _watchdog_task: Optional[asyncio.Task] = None
    _seconds_per_pat: float
    _monitored_names: dict[str, _MonitoredName]
    _pat_external_watchdog_process_args: list[str]

    def __init__(self, seconds_per_pat, services: ServicesInterface):
        super().__init__(KnownNames.watchdog_manager.value, services)
        self.lg = services.logger
        self._seconds_per_pat = seconds_per_pat
        self._monitored_names = dict()
        self._pat_external_watchdog_process_args = []

    def start(self):
        if self._watchdog_task is None:
            self._pat_external_watchdog_process_args = (
                self._services.get_external_watchdog_builder_class().pat_args(
                    str(self._services.settings.paths.name),
                )
            )
            self.lg.info(
                f"WatchdogManager: [{' '.join(self._pat_external_watchdog_process_args)}]"
            )
            now = time.time()
            for monitored in self._monitored_names.values():
                monitored.last_pat = now
            self._watchdog_task = asyncio.create_task(
                self._check_pats(), name="pat_watchdog"
            )

    def stop(self):
        if self._watchdog_task is not None and not self._watchdog_task.done():
            self._watchdog_task.cancel()

    async def join(self):
        if self._watchdog_task is not None:
            try:
                await self._watchdog_task
            except asyncio.CancelledError:
                pass

    def process_message(self, message: Message) -> None:
        # self.lg.path("++WatchdogManager.process_message")
        path_dbg = 0
        match message.Payload:
            case PatInternalWatchdog():
                path_dbg |= 0x00000001
                self._pat_internal_watchdog(message.src())
            case PatExternalWatchdog():
                path_dbg |= 0x00000002
                self._pat_external_watchdog()
            case _:
                path_dbg |= 0x00000004
                raise ValueError(
                    f"WatchdogManager does not handle message payloads of type {type(message.Payload)}"
                )
        # self.lg.path(f"--WatchdogManager.process_message  0x{path_dbg:08X}")

    def _pat_internal_watchdog(self, name: str):
        if name not in self._monitored_names:
            raise ValueError(
                f"ERROR. Received interal watchdog pat from unmonitored name: {name}. Monistored names: {list(self._monitored_names.keys())}"
            )
        self._monitored_names[name].last_pat = time.time()

    def add_monitored_name(self, monitored: MonitoredName) -> None:
        if monitored.timeout_seconds <= self._seconds_per_pat / 2:
            raise ValueError(
                f"ERROR. WatchdogManager cannot reliably monitor a timeout of {monitored.timeout_seconds} "
                f"(requested for {monitored.name}) because "
                f"WatchdogManager's _seconds_per_pat (sample rate) is {self._seconds_per_pat}"
            )
        if monitored.name in self._monitored_names:
            raise ValueError(
                f"ERROR. Name {monitored.name} is already being monitored with {self._monitored_names[monitored.name]}"
            )
        self._monitored_names[monitored.name] = _MonitoredName(
            monitored.name, monitored.timeout_seconds
        )

    def _check_for_timeout_expired(self) -> Optional[_MonitoredName]:
        # self.lg.path("++_check_for_timeout_expired")
        # path_dbg = 0
        expired: Optional[_MonitoredName] = None
        now = time.time()
        for monitored in self._monitored_names.values():
            # path_dbg |= 0x0000001
            required_pat_time = monitored.last_pat + monitored.timeout_seconds
            # self.lg.info(
            #     f"  {monitored.name:50s}  "
            #     f"{monitored.timeout_seconds:4d}  last_pat:{monitored.last_pat:11.1f}  "
            #     f"required_pat_time: {required_pat_time:11.1f}  "
            #     f"now:{now:11.1f}  "
            #     f"remaining: {int(required_pat_time - now):4d}  "
            #     f"required_pat_time < now ? {int(required_pat_time < now)}"
            # )
            if required_pat_time < now:
                # path_dbg |= 0x0000002
                expired = monitored
                break
        # self.lg.path(f"--_check_for_timeout_expired: count:{int(expired is not None)}  0x{path_dbg:08X}")
        return expired

    async def _check_pats(self) -> None:
        while (expired := self._check_for_timeout_expired()) is None:
            self._send(PatExternalWatchdogMessage())
            await asyncio.sleep(self._seconds_per_pat)
        self._send(
            InternalShutdownMessage(
                Src=self.name,
                Reason=(
                    f"Monitored object ({expired.name}) failed to pat internal watchdog.  \n"
                    f"  Last pat from {expired.name}: {int(time.time() - expired.last_pat)} seconds ago\n"
                    f"  Allowed seconds: {int(expired.timeout_seconds)}"
                ),
            )
        )

    def _pat_external_watchdog(self):
        if self._pat_external_watchdog_process_args:
            subprocess.run(self._pat_external_watchdog_process_args, check=True)



import argparse
import logging
import logging.config
import sys
import syslog
import traceback
from typing import Optional

from gwproactor.config.proactor_settings import ProactorSettings


def format_exceptions(exceptions: list[BaseException]) -> str:
    s = ""
    # noinspection PyBroadException
    try:
        if exceptions:
            for i, exception in enumerate(exceptions):
                s += f"++ Exception {i + 1:2d} / {len(exceptions)} ++++++++++++++++++++++++++++++++++++++++++++++++++++\n"
                try:
                    s += "".join(traceback.format_exception(exception))
                    s += "\n"
                except BaseException as e:
                    # noinspection PyBroadException
                    try:
                        s += f"ERROR formatting traceback for {e}\n"
                    except:
                        s += "UNEXPECTED ERROR formatting exception.\n"
                s += f"-- Exception {i + 1:2d} / {len(exceptions)} ----------------------------------------------------\n\n"
    except:
        s += "UNEXPECTED ERROR formatting exception.\n"
    return s


def setup_logging(
    args: argparse.Namespace,
    settings: ProactorSettings,
    errors: Optional[list[BaseException]] = None,
    add_screen_handler: bool = True,
    root_gets_handlers: bool = True,
) -> None:
    """Get python logging config based on parsed command line args, defaults, environment variables and logging config file.

    The order of precedence is:

        1. Command line arguments
        2. Environment
        3. Defaults

    """
    if errors is None:
        errors: list[BaseException] = []
    else:
        errors.clear()
    config_finished = False
    try:
        # Take any arguments from command line
        try:
            if getattr(args, "verbose", None):
                settings.logging.base_log_level = logging.INFO
                settings.logging.levels.message_summary = logging.DEBUG
            elif getattr(args, "message_summary", None):
                settings.logging.levels.message_summary = logging.INFO
        except BaseException as e:
            errors.append(e)

        # Create formatter from settings
        try:
            formatter = settings.logging.formatter.create()
        except BaseException as e:
            formatter = None
            errors.append(e)

        # Set application logger levels
        for logger_name, logger_settings in settings.logging.logger_levels().items():
            try:
                logger = logging.getLogger(logger_name)
                logger.setLevel(logger_settings["level"])
            except BaseException as e:
                errors.append(e)

        # Create handlers from settings, add them to root logger
        if root_gets_handlers:
            base_logger = logging.getLogger()
        else:
            base_logger = logging.getLogger(settings.logging.base_log_name)
            base_logger.propagate = False
        if add_screen_handler:
            try:
                # try not to add more than one screen handler.
                if not any(
                    [
                        h
                        for h in base_logger.handlers
                        if isinstance(h, logging.StreamHandler)
                        and (h.stream is sys.stderr or h.stream is sys.stdout)
                    ]
                ):
                    screen_handler = logging.StreamHandler()
                    if formatter is not None:
                        screen_handler.setFormatter(formatter)
                    base_logger.addHandler(screen_handler)
            except BaseException as e:
                errors.append(e)
        screen_handlers = [
            h
            for h in base_logger.handlers
            if isinstance(h, logging.StreamHandler)
            and (h.stream is sys.stderr or h.stream is sys.stdout)
        ]
        if len(screen_handlers) > 1:
            raise ValueError(
                "x More than 1 screen handlers  "
                f"{base_logger.name}  {len(screen_handlers)}  "
                f"\n\tstream handlers: {screen_handlers},  "
                f"\n\tstream handlers: {[type(x) for x in screen_handlers]},  "
                f"from all handlers {base_logger.handlers}"
            )
        try:
            file_handler = settings.logging.file_handler.create(
                settings.paths.log_dir, formatter
            )
            if formatter is not None:
                file_handler.setFormatter(formatter)
            base_logger.addHandler(file_handler)
        except BaseException as e:
            errors.append(e)
        config_finished = True
    except BaseException as e:
        config_finished = False
        errors.append(e)
    finally:
        # Try to tell user if something went wrong
        if errors:
            # noinspection PyBroadException
            try:
                s = "ERROR in setup_logging():\n" + format_exceptions(errors)
                if config_finished:
                    logging.error(s)
                else:
                    syslog.syslog(s)
                    print(s)
            except:
                pass
